@@ -24,6 +24,7 @@
  */
 #include <typeinfo>
 #include <algorithm>
+#include <utility>
 #include <boost/make_shared.hpp>
 #include <boost/date_time/posix_time/posix_time_types.hpp>
 #include <oxt/backtrace.hpp>
@@ -42,22 +43,11 @@ using namespace std;
 using namespace boost;
 using namespace oxt;
 
-
-template<typename T>
-static bool
-exceptionIsInstanceOf(const tracable_exception &e) {
-	try {
-		(void) dynamic_cast<T>(e);
-		return true;
-	} catch (const bad_cast &) {
-		return false;
-	}
-}
-
 #define TRY_COPY_EXCEPTION(klass) \
 	do { \
-		if (exceptionIsInstanceOf<const klass &>(e)) { \
-			return make_shared<klass>( (const klass &) e ); \
+		const klass *ep = dynamic_cast<const klass *>(&e); \
+		if (ep != NULL) { \
+			return boost::make_shared<klass>(*ep); \
 		} \
 	} while (false)
 
@@ -73,8 +63,9 @@ copyException(const tracable_exception &e) {
 	
 	TRY_COPY_EXCEPTION(ConfigurationException);
 	
-	TRY_COPY_EXCEPTION(SpawnException);
+	TRY_COPY_EXCEPTION(RequestQueueFullException);
 	TRY_COPY_EXCEPTION(GetAbortedException);
+	TRY_COPY_EXCEPTION(SpawnException);
 	
 	TRY_COPY_EXCEPTION(InvalidModeStringException);
 	TRY_COPY_EXCEPTION(ArgumentException);
@@ -91,13 +82,14 @@ copyException(const tracable_exception &e) {
 
 	TRY_COPY_EXCEPTION(boost::thread_interrupted);
 
-	return make_shared<tracable_exception>(e);
+	return boost::make_shared<tracable_exception>(e);
 }
 
 #define TRY_RETHROW_EXCEPTION(klass) \
 	do { \
-		if (exceptionIsInstanceOf<const klass &>(*e)) { \
-			throw klass((const klass &) *e); \
+		const klass *ep = dynamic_cast<const klass *>(&*e); \
+		if (ep != NULL) { \
+			throw klass(*ep); \
 		} \
 	} while (false)
 
@@ -114,6 +106,7 @@ rethrowException(const ExceptionPtr &e) {
 	TRY_RETHROW_EXCEPTION(ConfigurationException);
 	
 	TRY_RETHROW_EXCEPTION(SpawnException);
+	TRY_RETHROW_EXCEPTION(RequestQueueFullException);
 	TRY_RETHROW_EXCEPTION(GetAbortedException);
 	
 	TRY_RETHROW_EXCEPTION(InvalidModeStringException);
@@ -165,7 +158,24 @@ SuperGroup::generateSecret() const {
 }
 
 void
-SuperGroup::createInterruptableThread(const function<void ()> &func, const string &name,
+SuperGroup::runInitializationHooks() const {
+	getPool()->runHookScripts("after_initialize_supergroup",
+		boost::bind(&SuperGroup::setupInitializationOrDestructionHook, this, _1));
+}
+
+void
+SuperGroup::runDestructionHooks() const {
+	getPool()->runHookScripts("before_destroy_supergroup",
+		boost::bind(&SuperGroup::setupInitializationOrDestructionHook, this, _1));
+}
+
+void
+SuperGroup::setupInitializationOrDestructionHook(HookScriptOptions &options) const {
+	options.environment.push_back(make_pair("PASSENGER_APP_ROOT", this->options.appRoot));
+}
+
+void
+SuperGroup::createInterruptableThread(const boost::function<void ()> &func, const string &name,
 	unsigned int stackSize)
 {
 	getPool()->interruptableThreads.create_thread(func, name, stackSize);
@@ -187,7 +197,7 @@ SuperGroup::realDoInitialize(const Options &options, unsigned int generation) {
 		string message = "The directory " +
 			options.appRoot +
 			" does not seem to contain a web application.";
-		exception = make_shared<SpawnException>(
+		exception = boost::make_shared<SpawnException>(
 			message, message, false);
 	}
 	
@@ -201,7 +211,7 @@ SuperGroup::realDoInitialize(const Options &options, unsigned int generation) {
 			debug->messages->recv("Proceed with initializing SuperGroup");
 		}
 
-		unique_lock<boost::mutex> lock(getPoolSyncher(pool));
+		boost::unique_lock<boost::mutex> lock(getPoolSyncher(pool));
 		this_thread::disable_interruption di;
 		this_thread::disable_syscall_interruption dsi;
 		NOT_EXPECTING_EXCEPTIONS();
@@ -224,12 +234,12 @@ SuperGroup::realDoInitialize(const Options &options, unsigned int generation) {
 				const GetWaiter &waiter = getWaitlist.front();
 				actions.push_back(boost::bind(waiter.callback,
 					SessionPtr(), exception));
-				getWaitlist.pop();
+				getWaitlist.pop_front();
 			}
 		} else {
 			for (it = componentInfos.begin(); it != componentInfos.end(); it++) {
 				const ComponentInfo &info = *it;
-				GroupPtr group = make_shared<Group>(shared_from_this(),
+				GroupPtr group = boost::make_shared<Group>(shared_from_this(),
 					options, info);
 				groups.push_back(group);
 				if (info.isDefault) {
@@ -244,9 +254,11 @@ SuperGroup::realDoInitialize(const Options &options, unsigned int generation) {
 		verifyInvariants();
 		P_TRACE(2, "Done initializing SuperGroup " << inspect());
 	}
+
 	this_thread::disable_interruption di;
 	this_thread::disable_syscall_interruption dsi;
 	runAllActions(actions);
+	runInitializationHooks();
 }
 
 void
@@ -262,7 +274,7 @@ SuperGroup::realDoRestart(const Options &options, unsigned int generation) {
 		debug->messages->recv("Proceed with restarting SuperGroup");
 	}
 	
-	unique_lock<boost::mutex> lock(getPoolSyncher(pool));
+	boost::unique_lock<boost::mutex> lock(getPoolSyncher(pool));
 	if (OXT_UNLIKELY(this->generation != generation)) {
 		return;
 	}
@@ -292,7 +304,7 @@ SuperGroup::realDoRestart(const Options &options, unsigned int generation) {
 		} else {
 			// This is not an existing group but a new one,
 			// so create it.
-			group = make_shared<Group>(shared_from_this(),
+			group = boost::make_shared<Group>(shared_from_this(),
 				options, info);
 			newGroups.push_back(group);
 		}
@@ -333,12 +345,13 @@ Group::Group(const SuperGroupPtr &_superGroup, const Options &options, const Com
 	disabledCount  = 0;
 	spawner        = getPool()->spawnerFactory->create(options);
 	restartsInitiated = 0;
+	processesBeingSpawned = 0;
 	m_spawning     = false;
 	m_restarting   = false;
 	lifeStatus     = ALIVE;
 	if (options.restartDir.empty()) {
 		restartFile = options.appRoot + "/tmp/restart.txt";
-		alwaysRestartFile = options.appRoot + "/always_restart.txt";
+		alwaysRestartFile = options.appRoot + "/tmp/always_restart.txt";
 	} else if (options.restartDir[0] == '/') {
 		restartFile = options.restartDir + "/restart.txt";
 		alwaysRestartFile = options.restartDir + "/always_restart.txt";
@@ -372,7 +385,7 @@ Group::onSessionInitiateFailure(const ProcessPtr &process, Session *session) {
 	TRACE_POINT();
 	// Standard resource management boilerplate stuff...
 	PoolPtr pool = getPool();
-	unique_lock<boost::mutex> lock(pool->syncher);
+	boost::unique_lock<boost::mutex> lock(pool->syncher);
 	assert(process->isAlive());
 	assert(isAlive() || getLifeStatus() == SHUTTING_DOWN);
 
@@ -392,7 +405,7 @@ Group::onSessionClose(const ProcessPtr &process, Session *session) {
 	TRACE_POINT();
 	// Standard resource management boilerplate stuff...
 	PoolPtr pool = getPool();
-	unique_lock<boost::mutex> lock(pool->syncher);
+	boost::unique_lock<boost::mutex> lock(pool->syncher);
 	assert(process->isAlive());
 	assert(isAlive() || getLifeStatus() == SHUTTING_DOWN);
 
@@ -407,13 +420,13 @@ Group::onSessionClose(const ProcessPtr &process, Session *session) {
 		|| process->enabled == Process::DISABLING
 		|| process->enabled == Process::DETACHED);
 	if (process->enabled == Process::ENABLED) {
-		pqueue.decrease(process->pqHandle, process->utilization());
+		pqueue.decrease(process->pqHandle, process->busyness());
 	}
 
-	/* This group now has a process that's guaranteed to be not at
-	 * full utilization.
+	/* This group now has a process that's guaranteed to be not
+	 * totally busy.
 	 */
-	assert(!process->atFullUtilization());
+	assert(!process->isTotallyBusy());
 
 	bool detachingBecauseOfMaxRequests = false;
 	bool detachingBecauseCapacityNeeded = false;
@@ -447,8 +460,8 @@ Group::onSessionClose(const ProcessPtr &process, Session *session) {
 				 * checked conditions) then now's a good time to detach
 				 * this process or group in order to free capacity.
 				 */
-				P_DEBUG("Process " << process->inspect() << " is no longer at "
-					"full utilization; detaching it in order to make room in the pool");
+				P_DEBUG("Process " << process->inspect() << " is no longer totally "
+					"busy; detaching it in order to make room in the pool");
 			} else {
 				/* This process has processed its maximum number of requests,
 				 * so we detach it.
@@ -462,7 +475,7 @@ Group::onSessionClose(const ProcessPtr &process, Session *session) {
 			removeProcessFromList(process, disablingProcesses);
 			addProcessToList(process, disabledProcesses);
 			removeFromDisableWaitlist(process, DR_SUCCESS, actions);
-			asyncOOBWRequestIfNeeded(process);
+			maybeInitiateOobw(process);
 		}
 		
 		pool->fullVerifyInvariants();
@@ -471,7 +484,7 @@ Group::onSessionClose(const ProcessPtr &process, Session *session) {
 
 	} else {
 		// This could change process->enabled.
-		asyncOOBWRequestIfNeeded(process);
+		maybeInitiateOobw(process);
 
 		if (!getWaitlist.empty() && process->enabled == Process::ENABLED) {
 			/* If there are clients on this group waiting for a process to
@@ -488,53 +501,117 @@ void
 Group::requestOOBW(const ProcessPtr &process) {
 	// Standard resource management boilerplate stuff...
 	PoolPtr pool = getPool();
-	unique_lock<boost::mutex> lock(pool->syncher);
+	boost::unique_lock<boost::mutex> lock(pool->syncher);
 	if (isAlive() && process->isAlive() && process->oobwStatus == Process::OOBW_NOT_ACTIVE) {
 		process->oobwStatus = Process::OOBW_REQUESTED;
 	}
 }
 
-// The 'self' parameter is for keeping the current Group object alive
-void
-Group::lockAndAsyncOOBWRequestIfNeeded(const ProcessPtr &process, DisableResult result, GroupPtr self) {
-	TRACE_POINT();
-	
-	if (result != DR_SUCCESS && result != DR_CANCELED) {
-		return;
+bool
+Group::oobwAllowed() const {
+	unsigned int oobwInstances = 0;
+	foreach (const ProcessPtr &process, disablingProcesses) {
+		if (process->oobwStatus == Process::OOBW_IN_PROGRESS) {
+			oobwInstances += 1;
+		}
 	}
-	
-	// Standard resource management boilerplate stuff...
-	PoolPtr pool = getPool();
-	unique_lock<boost::mutex> lock(pool->syncher);
-	if (OXT_UNLIKELY(!process->isAlive() || !isAlive())) {
-		return;
+	foreach (const ProcessPtr &process, disabledProcesses) {
+		if (process->oobwStatus == Process::OOBW_IN_PROGRESS) {
+			oobwInstances += 1;
+		}
 	}
-	
-	P_DEBUG("Process " << process->inspect() << " disabled; proceeding with OOBW");
-	asyncOOBWRequestIfNeeded(process);
+	return oobwInstances < options.maxOutOfBandWorkInstances;
+}
+
+bool
+Group::shouldInitiateOobw(const ProcessPtr &process) const {
+	return process->oobwStatus == Process::OOBW_REQUESTED
+		&& process->enabled != Process::DETACHED
+		&& process->isAlive()
+		&& oobwAllowed();
 }
 
 void
-Group::asyncOOBWRequestIfNeeded(const ProcessPtr &process) {
-	if (process->oobwStatus != Process::OOBW_REQUESTED
-		|| process->enabled == Process::DETACHED
-		|| !process->isAlive())
-	{
+Group::maybeInitiateOobw(const ProcessPtr &process) {
+	if (shouldInitiateOobw(process)) {
+		initiateOobw(process);
+	}
+}
+
+// The 'self' parameter is for keeping the current Group object alive
+void
+Group::lockAndMaybeInitiateOobw(const ProcessPtr &process, DisableResult result, GroupPtr self) {
+	TRACE_POINT();
+	
+	// Standard resource management boilerplate stuff...
+	PoolPtr pool = getPool();
+	boost::unique_lock<boost::mutex> lock(pool->syncher);
+	if (OXT_UNLIKELY(!process->isAlive() || !isAlive())) {
 		return;
 	}
-	if (process->enabled == Process::ENABLED) {
+
+	assert(process->oobwStatus == Process::OOBW_IN_PROGRESS);
+
+	if (result == DR_SUCCESS) {
+		if (process->enabled == Process::DISABLED) {
+			P_DEBUG("Process " << process->inspect() << " disabled; proceeding " <<
+				"with out-of-band work");
+			process->oobwStatus = Process::OOBW_REQUESTED;
+			if (shouldInitiateOobw(process)) {
+				initiateOobw(process);
+			} else {
+				// We do not re-enable the process because it's likely that the
+				// administrator has explicitly changed the state.
+				P_DEBUG("Out-of-band work for process " << process->inspect() << " aborted "
+					"because the process no longer requests out-of-band work");
+				process->oobwStatus = Process::OOBW_NOT_ACTIVE;
+			}
+		} else {
+			// We do not re-enable the process because it's likely that the
+			// administrator has explicitly changed the state.
+			P_DEBUG("Out-of-band work for process " << process->inspect() << " aborted "
+				"because the process was reenabled after disabling");
+			process->oobwStatus = Process::OOBW_NOT_ACTIVE;
+		}
+	} else {
+		P_DEBUG("Out-of-band work for process " << process->inspect() << " aborted "
+			"because the process could not be disabled");
+		process->oobwStatus = Process::OOBW_NOT_ACTIVE;
+	}
+}
+
+void
+Group::initiateOobw(const ProcessPtr &process) {
+	assert(process->oobwStatus == Process::OOBW_REQUESTED);
+
+	process->oobwStatus = Process::OOBW_IN_PROGRESS;
+
+	if (process->enabled == Process::ENABLED
+	 || process->enabled == Process::DISABLING)
+	{
 		// We want the process to be disabled. However, disabling a process is potentially
 		// asynchronous, so we pass a callback which will re-aquire the lock and call this
 		// method again.
 		P_DEBUG("Disabling process " << process->inspect() << " in preparation for OOBW");
 		DisableResult result = disable(process,
-			boost::bind(&Group::lockAndAsyncOOBWRequestIfNeeded, this,
+			boost::bind(&Group::lockAndMaybeInitiateOobw, this,
 				_1, _2, shared_from_this()));
-		if (result == DR_DEFERRED) {
+		switch (result) {
+		case DR_SUCCESS:
+			// Continue code flow.
+			break;
+		case DR_DEFERRED:
+			// lockAndMaybeInitateOobw() will eventually be called.
 			return;
+		case DR_ERROR:
+		case DR_NOOP:
+			P_DEBUG("Out-of-band work for process " << process->inspect() << " aborted "
+				"because the process could not be disabled");
+			process->oobwStatus = Process::OOBW_NOT_ACTIVE;
+			return;
+		default:
+			P_BUG("Unexpected disable() result " << result);
 		}
-	} else if (process->enabled == Process::DISABLING) {
-		return;
 	}
 	
 	assert(process->enabled == Process::DISABLED);
@@ -543,7 +620,7 @@ Group::asyncOOBWRequestIfNeeded(const ProcessPtr &process) {
 	P_DEBUG("Initiating OOBW request for process " << process->inspect());
 	interruptableThreads.create_thread(
 		boost::bind(&Group::spawnThreadOOBWRequest, this, shared_from_this(), process),
-		"OOB request thread for process " + process->inspect(),
+		"OOBW request thread for process " + process->inspect(),
 		POOL_HELPER_THREAD_STACK_SIZE);
 }
 
@@ -569,7 +646,7 @@ Group::spawnThreadOOBWRequest(GroupPtr self, ProcessPtr process) {
 	UPDATE_TRACE_POINT();
 	{
 		// Standard resource management boilerplate stuff...
-		unique_lock<boost::mutex> lock(pool->syncher);
+		boost::unique_lock<boost::mutex> lock(pool->syncher);
 		if (OXT_UNLIKELY(!process->isAlive()
 			|| process->enabled == Process::DETACHED
 			|| !isAlive()))
@@ -587,7 +664,7 @@ Group::spawnThreadOOBWRequest(GroupPtr self, ProcessPtr process) {
 			return;
 		}
 		
-		assert(process->oobwStatus = Process::OOBW_IN_PROGRESS);
+		assert(process->oobwStatus == Process::OOBW_IN_PROGRESS);
 		assert(process->sessions == 0);
 		socket = process->sessionSockets.top();
 		assert(socket != NULL);
@@ -640,7 +717,7 @@ Group::spawnThreadOOBWRequest(GroupPtr self, ProcessPtr process) {
 	{
 		// Standard resource management boilerplate stuff...
 		PoolPtr pool = getPool();
-		unique_lock<boost::mutex> lock(pool->syncher);
+		boost::unique_lock<boost::mutex> lock(pool->syncher);
 		if (OXT_UNLIKELY(!process->isAlive() || !isAlive())) {
 			return;
 		}
@@ -650,8 +727,10 @@ Group::spawnThreadOOBWRequest(GroupPtr self, ProcessPtr process) {
 			enable(process, actions);
 			assignSessionsToGetWaiters(actions);
 		}
-		
+
 		pool->fullVerifyInvariants();
+
+		initiateNextOobwRequest();
 	}
 	UPDATE_TRACE_POINT();
 	runAllActions(actions);
@@ -664,6 +743,20 @@ Group::spawnThreadOOBWRequest(GroupPtr self, ProcessPtr process) {
 	}
 }
 
+void
+Group::initiateNextOobwRequest() {
+	ProcessList::const_iterator it, end = enabledProcesses.end();
+	for (it = enabledProcesses.begin(); it != end; it++) {
+		const ProcessPtr &process = *it;
+		if (shouldInitiateOobw(process)) {
+			// We keep an extra reference to processes to prevent premature destruction.
+			ProcessPtr p = process;
+			initiateOobw(p);
+			return;
+		}
+	}
+}
+
 // The 'self' parameter is for keeping the current Group object alive while this thread is running.
 void
 Group::spawnThreadMain(GroupPtr self, SpawnerPtr spawner, Options options, unsigned int restartsInitiated) {
@@ -726,7 +819,7 @@ Group::spawnThreadRealMain(const SpawnerPtr &spawner, const Options &options, un
 
 		UPDATE_TRACE_POINT();
 		ScopeGuard guard(boost::bind(Process::forceTriggerShutdownAndCleanup, process));
-		unique_lock<boost::mutex> lock(pool->syncher);
+		boost::unique_lock<boost::mutex> lock(pool->syncher);
 
 		if (!isAlive()) {
 			if (process != NULL) {
@@ -756,19 +849,31 @@ Group::spawnThreadRealMain(const SpawnerPtr &spawner, const Options &options, un
 
 		verifyInvariants();
 		assert(m_spawning);
+		assert(processesBeingSpawned > 0);
+
+		processesBeingSpawned--;
+		assert(processesBeingSpawned == 0);
 
 		UPDATE_TRACE_POINT();
 		vector<Callback> actions;
 		if (process != NULL) {
-			attach(process, actions);
-			guard.clear();
-			if (getWaitlist.empty()) {
-				pool->assignSessionsToGetWaiters(actions);
+			AttachResult result = attach(process, actions);
+			if (result == AR_OK) {
+				guard.clear();
+				if (getWaitlist.empty()) {
+					pool->assignSessionsToGetWaiters(actions);
+				} else {
+					assignSessionsToGetWaiters(actions);
+				}
+				P_DEBUG("New process count = " << enabledCount <<
+					", remaining get waiters = " << getWaitlist.size());
 			} else {
-				assignSessionsToGetWaiters(actions);
+				done = true;
+				P_DEBUG("Unable to attach spawned process " << process->inspect());
+				if (result == AR_ANOTHER_GROUP_IS_WAITING_FOR_CAPACITY) {
+					pool->possiblySpawnMoreProcessesForExistingGroups();
+				}
 			}
-			P_DEBUG("New process count = " << enabledCount <<
-				", remaining get waiters = " << getWaitlist.size());
 		} else {
 			// TODO: sure this is the best thing? if there are
 			// processes currently alive we should just use them.
@@ -783,21 +888,18 @@ Group::spawnThreadRealMain(const SpawnerPtr &spawner, const Options &options, un
 			done = true;
 		}
 
-		// Temporarily mark this Group as 'not spawning' so
-		// that pool->utilization() doesn't take this thread's spawning
-		// state into account.
-		m_spawning = false;
-		
 		done = done
-			|| ((unsigned long) enabledCount >= options.minProcesses && getWaitlist.empty())
+			|| (processLowerLimitsSatisfied() && getWaitlist.empty())
+			|| processUpperLimitsReached()
 			|| pool->atFullCapacity(false);
 		m_spawning = !done;
 		if (done) {
 			P_DEBUG("Spawn loop done");
 		} else {
+			processesBeingSpawned++;
 			P_DEBUG("Continue spawning");
 		}
-		
+
 		UPDATE_TRACE_POINT();
 		pool->fullVerifyInvariants();
 		lock.unlock();
@@ -813,13 +915,12 @@ Group::spawnThreadRealMain(const SpawnerPtr &spawner, const Options &options, un
 
 bool
 Group::shouldSpawn() const {
-	return !m_spawning
+	return allowSpawn()
 		&& (
-			(unsigned long) enabledCount < options.minProcesses
-			|| (enabledCount > 0 && pqueue.top()->atFullCapacity())
-		)
-		&& isAlive()
-		&& !poolAtFullCapacity();
+			!processLowerLimitsSatisfied()
+			|| allEnabledProcessesAreTotallyBusy()
+			|| !getWaitlist.empty()
+		);
 }
 
 bool
@@ -828,21 +929,24 @@ Group::shouldSpawnForGetAction() const {
 }
 
 void
-Group::restart(const Options &options) {
+Group::restart(const Options &options, RestartMethod method) {
 	vector<Callback> actions;
 
 	assert(isAlive());
-	assert(!m_restarting);
 	P_DEBUG("Restarting group " << name);
-	// Tell the restarter thread to exit as soon as possible.
+
+	// If there is currently a restarter thread or a spawner thread active,
+	// the following tells them to abort their current work as soon as possible.
 	restartsInitiated++;
-	m_spawning = false;
+
+	processesBeingSpawned = 0;
+	m_spawning   = false;
 	m_restarting = true;
 	detachAll(actions);
 	getPool()->interruptableThreads.create_thread(
 		boost::bind(&Group::finalizeRestart, this, shared_from_this(),
 			options.copyAndPersist().clearPerRequestFields(),
-			getPool()->spawnerFactory, actions),
+			method, getPool()->spawnerFactory, restartsInitiated, actions),
 		"Group restarter: " + name,
 		POOL_HELPER_THREAD_STACK_SIZE
 	);
@@ -850,7 +954,8 @@ Group::restart(const Options &options) {
 
 // The 'self' parameter is for keeping the current Group object alive while this thread is running.
 void
-Group::finalizeRestart(GroupPtr self, Options options, SpawnerFactoryPtr spawnerFactory,
+Group::finalizeRestart(GroupPtr self, Options options, RestartMethod method,
+	SpawnerFactoryPtr spawnerFactory, unsigned int restartsInitiated,
 	vector<Callback> postLockActions)
 {
 	TRACE_POINT();
@@ -882,6 +987,15 @@ Group::finalizeRestart(GroupPtr self, Options options, SpawnerFactoryPtr spawner
 		P_DEBUG("Group " << name << " is shutting down, so aborting restart");
 		return;
 	}
+	if (restartsInitiated != this->restartsInitiated) {
+		// Before this restart could be finalized, another restart command was given.
+		// The spawner we just created might be out of date now so we abort.
+		P_DEBUG("Restart of group " << name << " aborted because a new restart was initiated concurrently");
+		if (debug != NULL && debug->restarting) {
+			debug->debugger->send("Restarting aborted");
+		}
+		return;
+	}
 
 	// Run some sanity checks.
 	pool->fullVerifyInvariants();
@@ -894,13 +1008,23 @@ Group::finalizeRestart(GroupPtr self, Options options, SpawnerFactoryPtr spawner
 	spawner    = newSpawner;
 
 	m_restarting = false;
-	if (!getWaitlist.empty()) {
+	if (shouldSpawn()) {
 		spawn();
+	} else if (isWaitingForCapacity()) {
+		P_INFO("Group " << name << " is waiting for capacity to become available. "
+			"Trying to shutdown another idle process to free capacity...");
+		if (pool->forceFreeCapacity(this, postLockActions) != NULL) {
+			spawn();
+		} else {
+			P_INFO("There are no processes right now that are eligible "
+				"for shutdown. Will try again later.");
+		}
 	}
 	verifyInvariants();
 
 	l.unlock();
 	oldSpawner.reset();
+	Pool::runAllActions(postLockActions);
 	P_DEBUG("Restart of group " << name << " done");
 	if (debug != NULL && debug->restarting) {
 		debug->debugger->send("Restarting done");
@@ -931,8 +1055,14 @@ void
 Group::detachedProcessesCheckerMain(GroupPtr self) {
 	TRACE_POINT();
 	PoolPtr pool = getPool();
-	unique_lock<boost::mutex> lock(pool->syncher);
 
+	Pool::DebugSupportPtr debug = pool->debugSupport;
+	if (debug != NULL && debug->detachedProcessesChecker) {
+		debug->debugger->send("About to start detached processes checker");
+		debug->messages->recv("Proceed with starting detached processes checker");
+	}
+
+	boost::unique_lock<boost::mutex> lock(pool->syncher);
 	while (true) {
 		assert(detachedProcessesCheckerActive);
 
@@ -1028,21 +1158,40 @@ Group::poolAtFullCapacity() const {
 
 bool
 Group::anotherGroupIsWaitingForCapacity() const {
+	return findOtherGroupWaitingForCapacity() != NULL;
+}
+
+boost::shared_ptr<Group>
+Group::findOtherGroupWaitingForCapacity() const {
 	PoolPtr pool = getPool();
 	StringMap<SuperGroupPtr>::const_iterator sg_it, sg_end = pool->superGroups.end();
 	for (sg_it = pool->superGroups.begin(); sg_it != sg_end; sg_it++) {
 		pair<StaticString, SuperGroupPtr> p = *sg_it;
-		foreach (GroupPtr group, p.second->groups) {
-			if (group.get() != this
-			 && group->enabledProcesses.empty()
-			 && !group->spawning()
-			 && !group->getWaitlist.empty())
-			{
-				return true;
+		vector<GroupPtr>::const_iterator g_it, g_end = p.second->groups.end();
+		for (g_it = p.second->groups.begin(); g_it != g_end; g_it++) {
+			if (g_it->get() != this && (*g_it)->isWaitingForCapacity()) {
+				return *g_it;
 			}
 		}
 	}
-	return false;
+	return GroupPtr();
+}
+
+ProcessPtr
+Group::poolForceFreeCapacity(const Group *exclude, vector<Callback> &postLockActions) {
+	return getPool()->forceFreeCapacity(exclude, postLockActions);
+}
+
+bool
+Group::testOverflowRequestQueue() const {
+	// This has a performance penalty, although I'm not sure whether the penalty is
+	// any greater than a hash table lookup if I were to implement it in Options.
+	Pool::DebugSupportPtr debug = getPool()->debugSupport;
+	if (debug) {
+		return debug->testOverflowRequestQueue;
+	} else {
+		return false;
+	}
 }
 
 const ResourceLocator &
@@ -1050,6 +1199,26 @@ Group::getResourceLocator() const {
 	return getPool()->spawnerFactory->getResourceLocator();
 }
 
+// 'process' is not a reference so that bind(runAttachHooks, ...) causes the shared
+// pointer reference to increment.
+void
+Group::runAttachHooks(const ProcessPtr process) const {
+	getPool()->runHookScripts("attached_process",
+		boost::bind(&Group::setupAttachOrDetachHook, this, process, _1));
+}
+
+void
+Group::runDetachHooks(const ProcessPtr process) const {
+	getPool()->runHookScripts("detached_process",
+		boost::bind(&Group::setupAttachOrDetachHook, this, process, _1));
+}
+
+void
+Group::setupAttachOrDetachHook(const ProcessPtr process, HookScriptOptions &options) const {
+	options.environment.push_back(make_pair("PASSENGER_PROCESS_PID", toString(process->pid)));
+	options.environment.push_back(make_pair("PASSENGER_APP_ROOT", this->options.appRoot));
+}
+
 string
 Group::generateSecret(const SuperGroupPtr &superGroup) {
 	return superGroup->getPool()->randomGenerator->generateAsciiString(43);
@@ -1092,6 +1261,11 @@ Session::getGupid() const {
 	return getProcess()->gupid;
 }
 
+int
+Session::getStickySessionId() const {
+	return getProcess()->stickySessionId;
+}
+
 const GroupPtr
 Session::getGroup() const {
 	return getProcess()->getGroup();
@@ -1104,14 +1278,18 @@ Session::requestOOBW() {
 	process->getGroup()->requestOOBW(process);
 }
 
+int
+Session::kill(int signo) {
+	return getProcess()->kill(signo);
+}
+
 
 PipeWatcher::DataCallback PipeWatcher::onData;
 
-PipeWatcher::PipeWatcher(const FileDescriptor &_fd, const char *_name, pid_t _pid, bool _print)
+PipeWatcher::PipeWatcher(const FileDescriptor &_fd, const char *_name, pid_t _pid)
 	: fd(_fd),
 	  name(_name),
-	  pid(_pid),
-	  print(_print)
+	  pid(_pid)
 {
 	started = false;
 }
@@ -1131,7 +1309,7 @@ PipeWatcher::start() {
 }
 
 void
-PipeWatcher::threadMain(shared_ptr<PipeWatcher> self) {
+PipeWatcher::threadMain(boost::shared_ptr<PipeWatcher> self) {
 	TRACE_POINT();
 	self->threadMain();
 }
@@ -1140,7 +1318,7 @@ void
 PipeWatcher::threadMain() {
 	TRACE_POINT();
 	{
-		unique_lock<boost::mutex> lock(startSyncher);
+		boost::unique_lock<boost::mutex> lock(startSyncher);
 		while (!started) {
 			startCond.wait(lock);
 		}
@@ -1167,8 +1345,7 @@ PipeWatcher::threadMain() {
 			}
 		} else if (ret == 1 && buf[0] == '\n') {
 			UPDATE_TRACE_POINT();
-			P_LOG(print ? LVL_INFO : LVL_DEBUG,
-				"[App " << pid << " " << name << "] ");
+			printAppOutput(pid, name, "", 0);
 		} else {
 			UPDATE_TRACE_POINT();
 			vector<StaticString> lines;
@@ -1178,8 +1355,7 @@ PipeWatcher::threadMain() {
 			}
 			split(StaticString(buf, ret2), '\n', lines);
 			foreach (const StaticString line, lines) {
-				P_LOG(print ? LVL_INFO : LVL_DEBUG,
-					"[App " << pid << " " << name << "] " << line);
+				printAppOutput(pid, name, line.data(), line.size());
 			}
 		}
 
