@@ -101,6 +101,7 @@ private:
 		
 		result.push_back(&ruby);
 		result.push_back(&python);
+		result.push_back(&nodejs);
 		result.push_back(&loggingAgentAddress);
 		result.push_back(&loggingAgentUsername);
 		result.push_back(&loggingAgentPassword);
@@ -145,7 +146,10 @@ private:
 	}
 	
 public:
-	/*********** Spawn options that should be set manually ***********/
+	/*********** Spawn options that should be set manually ***********
+	 * These are the options that are relevant while spawning an application
+	 * process. These options are only used during spawning.
+	 */
 	
 	/**
 	 * The root directory of the application to spawn. In case of a Ruby on Rails
@@ -245,6 +249,12 @@ public:
 	 * is a Python app.
 	 */
 	StaticString python;
+
+	/**
+	 * Path to the Node.js command to use, in case the application to spawn
+	 * is a Node.js app.
+	 */
+	StaticString nodejs;
 	
 	/**
 	 * Any rights that the spawned application process may have. The SpawnManager
@@ -287,19 +297,58 @@ public:
 	bool raiseInternalError;
 	
 	
-	/*********** Per-group pool options that should be set manually ***********/
+	/*********** Per-group pool options that should be set manually ***********
+	 * These options dictate how Pool will manage processes, routing, etc. within
+	 * a single Group. These options are not process-specific, only group-specific.
+	 */
 	
 	/**
 	 * The minimum number of processes for the current group that the application
 	 * pool's cleaner thread should keep around.
 	 */
-	unsigned long minProcesses;
+	unsigned int minProcesses;
+
+	/**
+	 * The maximum number of processes that may be spawned
+	 * for this app root. This option only has effect if it's lower than
+	 * the pool size.
+	 *
+	 * A value of 0 means unspecified, and has no effect.
+	 */
+	unsigned int maxProcesses;
 	
 	/** The number of seconds that preloader processes may stay alive idling. */
 	long maxPreloaderIdleTime;
+
+	/**
+	 * The maximum number of processes inside a group that may be performing
+	 * out-of-band work at the same time.
+	 */
+	unsigned int maxOutOfBandWorkInstances;
+
+	/**
+	 * The maximum number of requests that may live in the Group.getWaitlist queue.
+	 * A value of 0 means unlimited.
+	 */
+	unsigned int maxRequestQueueSize;
+
+	/**
+	 * The Union Station key to use in case analytics logging is enabled.
+	 * It is used by Pool::collectAnalytics() and other administrative
+	 * functions which are called periodically. Because they do not belong
+	 * to any request, and they may still want to log to Union Station,
+	 * this key is stored in the per-group options structure.
+	 *
+	 * It is not used on a per-request basis. Per-request analytics logging
+	 * (and Union Station logging) uses the logger object in the `logger` field
+	 * instead.
+	 */
+	StaticString unionStationKey;
+
+	/*-----------------*/
 	
 	
-	/*********** Per-request options that should be set manually and that only matter to Pool ***********/
+	/*********** Per-request options that should be set manually ***********/
 	
 	/** Current request host name. */
 	StaticString hostName;
@@ -314,9 +363,9 @@ public:
 	UnionStation::LoggerPtr logger;
 
 	/**
-	 * The Union Station key to use in case analytics logging is enabled.
+	 * A sticky session ID for routing to a specific process.
 	 */
-	StaticString unionStationKey;
+	unsigned int stickySessionId;
 	
 	/**
 	 * A throttling rate for file stats. When set to a non-zero value N,
@@ -342,16 +391,14 @@ public:
 	 */
 	bool noop;
 
-	/** Specifies whether, if the pool is already full, the pool is allowed to
-	 * trash a non-idle process in order to free capacity. True by default.
+	/*-----------------*/
+	/*-----------------*/
+	
+	
+	/*********** Spawn options automatically set by Pool ***********
+	 * These options are passed to the Spawner. The Pool::get() caller may not
+	 * see these values.
 	 */
-	bool allowTrashingNonIdleProcesses;
-
-	/*-----------------*/
-	/*-----------------*/
-	
-	
-	/*********** Spawn options automatically set by Pool ***********/
 	
 	/** The secret key of the pool group that the spawned process is to belong to. */
 	StaticString groupSecret;
@@ -370,8 +417,9 @@ public:
 		baseURI                 = "/";
 		spawnMethod             = "smart";
 		defaultUser             = "nobody";
-		ruby                    = "ruby";
-		python                  = "python";
+		ruby                    = DEFAULT_RUBY;
+		python                  = DEFAULT_PYTHON;
+		nodejs                  = DEFAULT_NODEJS;
 		rights                  = DEFAULT_BACKEND_ACCOUNT_RIGHTS;
 		debugger                = false;
 		loadShellEnvvars        = true;
@@ -379,13 +427,16 @@ public:
 		raiseInternalError      = false;
 		
 		minProcesses            = 1;
+		maxProcesses            = 0;
 		maxPreloaderIdleTime    = -1;
+		maxOutOfBandWorkInstances = 1;
+		maxRequestQueueSize     = 100;
 		
+		stickySessionId         = 0;
 		statThrottleRate        = 0;
 		maxRequests             = 0;
 		noop                    = false;
-		allowTrashingNonIdleProcesses = true;
-		
+
 		/*********************************/
 	}
 	
@@ -470,8 +521,9 @@ public:
 	}
 	
 	Options &clearPerRequestFields() {
-		hostName = string();
-		uri      = string();
+		hostName = StaticString();
+		uri      = StaticString();
+		stickySessionId = 0;
 		noop     = false;
 		return clearLogger();
 	}
@@ -516,18 +568,21 @@ public:
 			appendKeyValue (vec, "postexec_chroot",    postexecChroot);
 			appendKeyValue (vec, "ruby",               ruby);
 			appendKeyValue (vec, "python",             python);
+			appendKeyValue (vec, "nodejs",             nodejs);
 			appendKeyValue (vec, "logging_agent_address",  loggingAgentAddress);
 			appendKeyValue (vec, "logging_agent_username", loggingAgentUsername);
 			appendKeyValue (vec, "logging_agent_password", loggingAgentPassword);
 			appendKeyValue4(vec, "debugger",           debugger);
 			appendKeyValue4(vec, "analytics",          analytics);
-			appendKeyValue (vec, "union_station_key",  unionStationKey);
 
 			appendKeyValue (vec, "group_secret",       groupSecret);
 		}
 		if (fields & PER_GROUP_POOL_OPTIONS) {
 			appendKeyValue3(vec, "min_processes",       minProcesses);
+			appendKeyValue3(vec, "max_processes",       maxProcesses);
 			appendKeyValue2(vec, "max_preloader_idle_time", maxPreloaderIdleTime);
+			appendKeyValue3(vec, "max_out_of_band_work_instances", maxOutOfBandWorkInstances);
+			appendKeyValue (vec, "union_station_key",   unionStationKey);
 		}
 		
 		/*********************************/
@@ -567,17 +622,25 @@ public:
 			return ruby + "\t" + resourceLocator.getHelperScriptsDir() + "/rack-loader.rb";
 		} else if (appType == "wsgi") {
 			return python + "\t" + resourceLocator.getHelperScriptsDir() + "/wsgi-loader.py";
+		} else if (appType == "node") {
+			return nodejs + "\t" + resourceLocator.getHelperScriptsDir() + "/node-loader.js";
+		} else if (appType == "meteor") {
+			return ruby + "\t" + resourceLocator.getHelperScriptsDir() + "/meteor-loader.rb";
 		} else {
 			return startCommand;
 		}
 	}
 	
 	StaticString getStartupFile() const {
-		const char *result = getAppTypeStartupFile(getAppType(appType));
-		if (result == NULL) {
-			return startupFile;
+		if (startupFile.empty()) {
+			const char *result = getAppTypeStartupFile(getAppType(appType));
+			if (result == NULL) {
+				return "";
+			} else {
+				return result;
+			}
 		} else {
-			return result;
+			return startupFile;
 		}
 	}
 	
