@@ -27,6 +27,7 @@
 
 #include <string>
 #include <vector>
+#include <algorithm>
 #include <utility>
 #include <sstream>
 #include <iomanip>
@@ -48,9 +49,9 @@
 #include <ApplicationPool2/Options.h>
 #include <UnionStation.h>
 #include <Logging.h>
-#include <SafeLibev.h>
 #include <Exceptions.h>
 #include <RandomGenerator.h>
+#include <Hooks.h>
 #include <Utils/Lock.h>
 #include <Utils/AnsiColorConstants.h>
 #include <Utils/SystemTime.h>
@@ -66,7 +67,7 @@ using namespace boost;
 using namespace oxt;
 
 
-class Pool: public enable_shared_from_this<Pool> {
+class Pool: public boost::enable_shared_from_this<Pool> {
 public:
 	struct InspectOptions {
 		bool colorize;
@@ -92,7 +93,9 @@ public:
 	typedef UnionStation::LoggerPtr LoggerPtr;
 
 	struct DebugSupport {
+		/** Mailbox for the unit tests to receive messages on. */
 		MessageBoxPtr debugger;
+		/** Mailbox for the ApplicationPool code to receive messages on. */
 		MessageBoxPtr messages;
 
 		// Choose aspects to debug.
@@ -100,34 +103,37 @@ public:
 		bool spawning;
 		bool superGroup;
 		bool oobw;
+		bool testOverflowRequestQueue;
+		bool detachedProcessesChecker;
 
 		// The following fields may only be accessed by Pool.
 		boost::mutex syncher;
 		unsigned int spawnLoopIteration;
 
 		DebugSupport() {
-			debugger = make_shared<MessageBox>();
-			messages = make_shared<MessageBox>();
+			debugger = boost::make_shared<MessageBox>();
+			messages = boost::make_shared<MessageBox>();
 			restarting = true;
 			spawning   = true;
 			superGroup = false;
 			oobw       = false;
+			detachedProcessesChecker = false;
+			testOverflowRequestQueue = false;
 			spawnLoopIteration = 0;
 		}
 	};
 
-	typedef shared_ptr<DebugSupport> DebugSupportPtr;
+	typedef boost::shared_ptr<DebugSupport> DebugSupportPtr;
 
 	SpawnerFactoryPtr spawnerFactory;
 	LoggerFactoryPtr loggerFactory;
 	RandomGeneratorPtr randomGenerator;
 
 	mutable boost::mutex syncher;
-	SafeLibev *libev;
 	unsigned int max;
 	unsigned long long maxIdleTime;
 	
-	condition_variable garbageCollectionCond;
+	boost::condition_variable garbageCollectionCond;
 	
 	/**
 	 * Code can register background threads in one of these dynamic thread groups
@@ -160,7 +166,7 @@ public:
 	 *
 	 * - A process has been spawned but its associated group has
 	 *   no get waiters. This process can be killed and the resulting
-	 *   free capacity will be used to use spawn a process for this
+	 *   free capacity will be used to spawn a process for this
 	 *   get request.
 	 * - A process (that has apparently been spawned after getWaitlist
 	 *   was populated) is done processing a request. This process can
@@ -187,6 +193,7 @@ public:
 	 */
 	vector<GetWaiter> getWaitlist;
 
+	const VariantMap *agentsOptions;
 	DebugSupportPtr debugSupport;
 	
 	static void runAllActions(const vector<Callback> &actions) {
@@ -241,7 +248,36 @@ public:
 		}
 	}
 	
-	ProcessPtr findOldestIdleProcess() const {
+	bool runHookScripts(const char *name,
+		const boost::function<void (HookScriptOptions &)> &setup) const
+	{
+		if (agentsOptions != NULL) {
+			string hookName = string("hook_") + name;
+			string spec = agentsOptions->get(hookName, false);
+			if (!spec.empty()) {
+				HookScriptOptions options;
+				options.agentsOptions = agentsOptions;
+				options.name = name;
+				options.spec = spec;
+				setup(options);
+				return Passenger::runHookScripts(options);
+			} else {
+				return true;
+			}
+		} else {
+			return true;
+		}
+	}
+
+	static const char *maybePluralize(unsigned int count, const char *singular, const char *plural) {
+		if (count == 1) {
+			return singular;
+		} else {
+			return plural;
+		}
+	}
+
+	ProcessPtr findOldestIdleProcess(const Group *exclude = NULL) const {
 		ProcessPtr oldestIdleProcess;
 		
 		SuperGroupMap::const_iterator it, end = superGroups.end();
@@ -251,11 +287,14 @@ public:
 			vector<GroupPtr>::const_iterator g_it, g_end = groups.end();
 			for (g_it = groups.begin(); g_it != g_end; g_it++) {
 				const GroupPtr &group = *g_it;
+				if (group.get() == exclude) {
+					continue;
+				}
 				const ProcessList &processes = group->enabledProcesses;
 				ProcessList::const_iterator p_it, p_end = processes.end();
 				for (p_it = processes.begin(); p_it != p_end; p_it++) {
 					const ProcessPtr process = *p_it;
-					if (process->utilization() == 0
+					if (process->busyness() == 0
 					     && (oldestIdleProcess == NULL
 					         || process->lastUsed < oldestIdleProcess->lastUsed)
 					) {
@@ -301,13 +340,14 @@ public:
 		bool done = false;
 		vector<GetWaiter>::iterator it, end = getWaitlist.end();
 		vector<GetWaiter> newWaitlist;
-		
+
 		for (it = getWaitlist.begin(); it != end && !done; it++) {
 			GetWaiter &waiter = *it;
-			
+
 			SuperGroup *superGroup = findMatchingSuperGroup(waiter.options);
 			if (superGroup != NULL) {
-				SessionPtr session = superGroup->get(waiter.options, waiter.callback);
+				SessionPtr session = superGroup->get(waiter.options, waiter.callback,
+					postLockActions);
 				if (session != NULL) {
 					postLockActions.push_back(boost::bind(
 						waiter.callback, session, ExceptionPtr()));
@@ -316,7 +356,8 @@ public:
 				 *       the group's get wait list.
 				 */
 			} else if (!atFullCapacity(false)) {
-				createSuperGroupAndAsyncGetFromIt(waiter.options, waiter.callback);
+				createSuperGroupAndAsyncGetFromIt(waiter.options, waiter.callback,
+					postLockActions);
 			} else {
 				/* Still cannot satisfy this get request. Keep it on the get
 				 * wait list and try again later.
@@ -324,8 +365,8 @@ public:
 				newWaitlist.push_back(waiter);
 			}
 		}
-		
-		getWaitlist = newWaitlist;
+
+		std::swap(getWaitlist, newWaitlist);
 	}
 
 	template<typename Queue>
@@ -337,7 +378,7 @@ public:
 			postLockActions.push_back(boost::bind(
 				getWaitlist.front().callback, SessionPtr(),
 				exception));
-			getWaitlist.pop();
+			getWaitlist.pop_front();
 		}
 	}
 	
@@ -379,9 +420,33 @@ public:
 		getWaitlist.reserve(getWaitlist.size() + superGroup->getWaitlist.size());
 		while (!superGroup->getWaitlist.empty()) {
 			getWaitlist.push_back(superGroup->getWaitlist.front());
-			superGroup->getWaitlist.pop();
+			superGroup->getWaitlist.pop_front();
 		}
 	}
+
+	/**
+	 * Calls Group::detach() so be sure to fix up the invariants afterwards.
+	 * See the comments for Group::detach() and the code for detachProcessUnlocked().
+	 */
+	ProcessPtr forceFreeCapacity(const Group *exclude,
+		vector<Callback> &postLockActions)
+	{
+		ProcessPtr process = findOldestIdleProcess(exclude);
+		if (process != NULL) {
+			P_DEBUG("Forcefully detaching process " << process->inspect() <<
+				" in order to free capacity in the pool");
+
+			const GroupPtr group = process->getGroup();
+			assert(group != NULL);
+			assert(group->getWaitlist.empty());
+
+			const SuperGroupPtr superGroup = group->getSuperGroup();
+			assert(superGroup != NULL);
+
+			group->detach(process, postLockActions);
+		}
+		return process;
+	}
 	
 	/**
 	 * Forcefully destroys and detaches the given SuperGroup. After detaching
@@ -410,7 +475,7 @@ public:
 			const SuperGroupPtr superGroup = group->getSuperGroup();
 			assert(superGroup->state != SuperGroup::INITIALIZING);
 			assert(superGroup->getWaitlist.empty());
-			
+
 			group->detach(process, postLockActions);
 			// 'process' may now be a stale pointer so don't use it anymore.
 			assignSessionsToGetWaiters(postLockActions);
@@ -470,7 +535,7 @@ public:
 
 	struct DetachSuperGroupWaitTicket {
 		boost::mutex syncher;
-		condition_variable cond;
+		boost::condition_variable cond;
 		SuperGroup::ShutdownResult result;
 		bool done;
 
@@ -481,7 +546,7 @@ public:
 
 	struct DisableWaitTicket {
 		boost::mutex syncher;
-		condition_variable cond;
+		boost::condition_variable cond;
 		DisableResult result;
 		bool done;
 
@@ -491,7 +556,7 @@ public:
 	};
 
 	static void syncDetachSuperGroupCallback(SuperGroup::ShutdownResult result,
-		shared_ptr<DetachSuperGroupWaitTicket> ticket)
+		boost::shared_ptr<DetachSuperGroupWaitTicket> ticket)
 	{
 		LockGuard l(ticket->syncher);
 		ticket->done = true;
@@ -499,7 +564,7 @@ public:
 		ticket->cond.notify_one();
 	}
 
-	static void waitDetachSuperGroupCallback(shared_ptr<DetachSuperGroupWaitTicket> ticket) {
+	static void waitDetachSuperGroupCallback(boost::shared_ptr<DetachSuperGroupWaitTicket> ticket) {
 		ScopedLock l(ticket->syncher);
 		while (!ticket->done) {
 			ticket->cond.wait(l);
@@ -507,7 +572,7 @@ public:
 	}
 
 	static void syncDisableProcessCallback(const ProcessPtr &process, DisableResult result,
-		shared_ptr<DisableWaitTicket> ticket)
+		boost::shared_ptr<DisableWaitTicket> ticket)
 	{
 		LockGuard l(ticket->syncher);
 		ticket->done = true;
@@ -571,7 +636,8 @@ public:
 			process->lastUsed + maxIdleTime;
 		if (process->sessions == 0
 		 && state.now >= processGcTime
-		 && (unsigned long) group->enabledCount > group->options.minProcesses) {
+		 && (unsigned long) group->getProcessCount() > group->options.minProcesses)
+		{
 			ProcessList::iterator prev = p_it;
 			prev--;
 			P_DEBUG("Garbage collect idle process: " << process->inspect() <<
@@ -669,7 +735,7 @@ public:
 		stringstream data;
 	};
 
-	typedef shared_ptr<ProcessAnalyticsLogEntry> ProcessAnalyticsLogEntryPtr;
+	typedef boost::shared_ptr<ProcessAnalyticsLogEntry> ProcessAnalyticsLogEntryPtr;
 
 	static void collectAnalytics(PoolPtr self) {
 		TRACE_POINT();
@@ -779,7 +845,7 @@ public:
 
 					// Log to Union Station.
 					if (group->options.analytics && loggerFactory != NULL) {
-						ProcessAnalyticsLogEntryPtr entry = make_shared<ProcessAnalyticsLogEntry>();
+						ProcessAnalyticsLogEntryPtr entry = boost::make_shared<ProcessAnalyticsLogEntry>();
 						stringstream &xml = entry->data;
 
 						entry->groupName = group->name;
@@ -829,7 +895,7 @@ public:
 	}
 	
 	SuperGroupPtr createSuperGroup(const Options &options) {
-		SuperGroupPtr superGroup = make_shared<SuperGroup>(shared_from_this(),
+		SuperGroupPtr superGroup = boost::make_shared<SuperGroup>(shared_from_this(),
 			options);
 		superGroup->initialize();
 		superGroups.set(options.getAppGroupName(), superGroup);
@@ -838,10 +904,11 @@ public:
 	}
 	
 	SuperGroupPtr createSuperGroupAndAsyncGetFromIt(const Options &options,
-		const GetCallback &callback)
+		const GetCallback &callback, vector<Callback> &postLockActions)
 	{
 		SuperGroupPtr superGroup = createSuperGroup(options);
-		SessionPtr session = superGroup->get(options, callback);
+		SessionPtr session = superGroup->get(options, callback,
+			postLockActions);
 		/* Callback should now have been put on the wait list,
 		 * unless something has changed and we forgot to update
 		 * some code here...
@@ -854,18 +921,19 @@ public:
 	const SuperGroupPtr getSuperGroup(const char *name);
 	
 public:
-	Pool(SafeLibev *libev, const SpawnerFactoryPtr &spawnerFactory,
+	Pool(const SpawnerFactoryPtr &spawnerFactory,
 		const LoggerFactoryPtr &loggerFactory = LoggerFactoryPtr(),
-		const RandomGeneratorPtr &randomGenerator = RandomGeneratorPtr())
+		const RandomGeneratorPtr &randomGenerator = RandomGeneratorPtr(),
+		const VariantMap *agentsOptions = NULL)
 	{
-		this->libev = libev;
 		this->spawnerFactory = spawnerFactory;
 		this->loggerFactory = loggerFactory;
 		if (randomGenerator != NULL) {
 			this->randomGenerator = randomGenerator;
 		} else {
-			this->randomGenerator = make_shared<RandomGenerator>();
+			this->randomGenerator = boost::make_shared<RandomGenerator>();
 		}
+		this->agentsOptions = agentsOptions;
 		
 		lifeStatus  = ALIVE;
 		max         = 6;
@@ -901,7 +969,7 @@ public:
 
 	void initDebugging() {
 		LockGuard l(syncher);
-		debugSupport = make_shared<DebugSupport>();
+		debugSupport = boost::make_shared<DebugSupport>();
 	}
 
 	void destroy() {
@@ -935,17 +1003,18 @@ public:
 	// should never call the callback while holding the lock.
 	void asyncGet(const Options &options, const GetCallback &callback, bool lockNow = true) {
 		DynamicScopedLock lock(syncher, lockNow);
-		
+
 		assert(lifeStatus == ALIVE);
 		verifyInvariants();
-		P_TRACE(2, "asyncGet(appRoot=" << options.appRoot << ")");
-		
+		P_TRACE(2, "asyncGet(appGroupName=" << options.getAppGroupName() << ")");
+		vector<Callback> actions;
+
 		SuperGroup *existingSuperGroup = findMatchingSuperGroup(options);
 		if (OXT_LIKELY(existingSuperGroup != NULL)) {
 			/* Best case: the app super group is already in the pool. Let's use it. */
 			P_TRACE(2, "Found existing SuperGroup");
 			existingSuperGroup->verifyInvariants();
-			SessionPtr session = existingSuperGroup->get(options, callback);
+			SessionPtr session = existingSuperGroup->get(options, callback, actions);
 			existingSuperGroup->verifyInvariants();
 			verifyInvariants();
 			P_TRACE(2, "asyncGet() finished");
@@ -961,36 +1030,21 @@ public:
 			 * resources to make a new one.
 			 */
 			P_DEBUG("Spawning new SuperGroup");
-			SuperGroupPtr superGroup = createSuperGroupAndAsyncGetFromIt(options, callback);
+			SuperGroupPtr superGroup = createSuperGroupAndAsyncGetFromIt(options,
+				callback, actions);
 			superGroup->verifyInvariants();
 			verifyInvariants();
 			P_DEBUG("asyncGet() finished");
 			
 		} else {
-			vector<Callback> actions;
-			
 			/* Uh oh, the app super group isn't in the pool but we don't
 			 * have the resources to make a new one. The sysadmin should
 			 * configure the system to let something like this happen
 			 * as least as possible, but let's try to handle it as well
 			 * as we can.
-			 *
-			 * First, try to trash an idle process that's the oldest.
 			 */
-			P_DEBUG("Pool is at full capacity; trying to free a process...");
-			ProcessPtr process = findOldestIdleProcess();
-			if (process == NULL) {
-				/* All processes are doing something. We have no choice
-				 * but to trash a non-idle process.
-				 */
-				if (options.allowTrashingNonIdleProcesses) {
-					process = findBestProcessToTrash();
-				}
-			} else {
-				// Check invariant.
-				assert(process->getGroup()->getWaitlist.empty());
-			}
-			if (process == NULL) {
+			ProcessPtr freedProcess = forceFreeCapacity(NULL, actions);
+			if (freedProcess == NULL) {
 				/* No process is eligible for killing. This could happen if, for example,
 				 * all (super)groups are currently initializing/restarting/spawning/etc.
 				 * We have no choice but to satisfy this get() action later when resources
@@ -1001,33 +1055,24 @@ public:
 					options.copyAndPersist().clearLogger(),
 					callback));
 			} else {
-				GroupPtr group;
-				SuperGroupPtr superGroup;
-				
-				P_DEBUG("Freeing process " << process->inspect());
-				group = process->getGroup();
-				assert(group != NULL);
-				superGroup = group->getSuperGroup();
-				assert(superGroup != NULL);
-				
-				group->detach(process, actions);
-				
 				/* Now that a process has been trashed we can create
 				 * the missing SuperGroup.
 				 */
 				P_DEBUG("Creating new SuperGroup");
-				superGroup = make_shared<SuperGroup>(shared_from_this(), options);
+				SuperGroupPtr superGroup;
+				superGroup = boost::make_shared<SuperGroup>(shared_from_this(), options);
 				superGroup->initialize();
 				superGroups.set(options.getAppGroupName(), superGroup);
 				garbageCollectionCond.notify_all();
-				SessionPtr session = superGroup->get(options, callback);
+				SessionPtr session = superGroup->get(options, callback,
+					actions);
 				/* The SuperGroup is still initializing so the callback
 				 * should now have been put on the wait list,
 				 * unless something has changed and we forgot to update
 				 * some code here...
 				 */
 				assert(session == NULL);
-				group->verifyInvariants();
+				freedProcess->getGroup()->verifyInvariants();
 				superGroup->verifyInvariants();
 			}
 			
@@ -1035,22 +1080,24 @@ public:
 			verifyInvariants();
 			verifyExpensiveInvariants();
 			P_TRACE(2, "asyncGet() finished");
-			
-			if (!actions.empty()) {
-				if (lockNow) {
+		}
+
+		if (!actions.empty()) {
+			if (lockNow) {
+				if (lock.owns_lock()) {
 					lock.unlock();
-					runAllActions(actions);
-				} else {
-					// This state is not allowed. If we reach
-					// here then it probably indicates a bug in
-					// the test suite.
-					abort();
 				}
+				runAllActions(actions);
+			} else {
+				// This state is not allowed. If we reach
+				// here then it probably indicates a bug in
+				// the test suite.
+				abort();
 			}
 		}
 	}
 	
-	// TODO: 'ticket' should be a shared_ptr for interruption-safety.
+	// TODO: 'ticket' should be a boost::shared_ptr for interruption-safety.
 	SessionPtr get(const Options &options, Ticket *ticket) {
 		ticket->session.reset();
 		ticket->exception.reset();
@@ -1123,20 +1170,20 @@ public:
 		garbageCollectionCond.notify_all();
 	}
 	
-	unsigned int utilization(bool lock = true) const {
+	unsigned int capacityUsed(bool lock = true) const {
 		DynamicScopedLock l(syncher, lock);
 		SuperGroupMap::const_iterator it, end = superGroups.end();
 		int result = 0;
 		for (it = superGroups.begin(); it != end; it++) {
 			const SuperGroupPtr &superGroup = it->second;
-			result += superGroup->utilization();
+			result += superGroup->capacityUsed();
 		}
 		return result;
 	}
 	
 	bool atFullCapacity(bool lock = true) const {
 		DynamicScopedLock l(syncher, lock);
-		return utilization(false) >= max;
+		return capacityUsed(false) >= max;
 	}
 
 	vector<ProcessPtr> getProcesses(bool lock = true) const {
@@ -1165,6 +1212,11 @@ public:
 		return result;
 	}
 	
+	/**
+	 * Returns the total number of processes in the pool, including all disabling and
+	 * disabled processes, but excluding processes that are shutting down and excluding
+	 * processes that are being spawned.
+	 */
 	unsigned int getProcessCount(bool lock = true) const {
 		DynamicScopedLock l(syncher, lock);
 		unsigned int result = 0;
@@ -1229,10 +1281,10 @@ public:
 				verifyExpensiveInvariants();
 				
 				vector<Callback> actions;
-				shared_ptr<DetachSuperGroupWaitTicket> ticket =
-					make_shared<DetachSuperGroupWaitTicket>();
+				boost::shared_ptr<DetachSuperGroupWaitTicket> ticket =
+					boost::make_shared<DetachSuperGroupWaitTicket>();
 				ExceptionPtr exception = copyException(
-					GetAbortedException("The containg SuperGroup was detached."));
+					GetAbortedException("The containing SuperGroup was detached."));
 
 				forceDetachSuperGroup(superGroup, actions,
 					boost::bind(syncDetachSuperGroupCallback, _1, ticket));
@@ -1331,8 +1383,8 @@ public:
 		ProcessPtr process = findProcessByGupid(gupid, false);
 		if (process != NULL) {
 			GroupPtr group = process->getGroup();
-			// Must be a shared_ptr to be interruption-safe.
-			shared_ptr<DisableWaitTicket> ticket = make_shared<DisableWaitTicket>();
+			// Must be a boost::shared_ptr to be interruption-safe.
+			boost::shared_ptr<DisableWaitTicket> ticket = boost::make_shared<DisableWaitTicket>();
 			DisableResult result = group->disable(process,
 				boost::bind(syncDisableProcessCallback, _1, _2, ticket));
 			group->verifyInvariants();
@@ -1352,34 +1404,33 @@ public:
 		}
 	}
 
-	unsigned int restartGroupsByAppRoot(const string &appRoot) {
+	bool restartGroupByName(const StaticString &name, RestartMethod method = RM_DEFAULT) {
 		ScopedLock l(syncher);
 		SuperGroupMap::iterator sg_it, sg_end = superGroups.end();
-		unsigned int result = 0;
 
 		for (sg_it = superGroups.begin(); sg_it != sg_end; sg_it++) {
 			const SuperGroupPtr &superGroup = sg_it->second;
 			foreach (const GroupPtr &group, superGroup->groups) {
-				if (group->options.appRoot == appRoot) {
-					result++;
+				if (name == group->name) {
 					if (!group->restarting()) {
-						group->restart(group->options);
+						group->restart(group->options, method);
 					}
+					return true;
 				}
 			}
 		}
 
-		return result;
+		return false;
 	}
 
-	unsigned int restartSuperGroupsByAppRoot(const string &appRoot) {
+	unsigned int restartSuperGroupsByAppRoot(const StaticString &appRoot) {
 		ScopedLock l(syncher);
 		SuperGroupMap::iterator sg_it, sg_end = superGroups.end();
 		unsigned int result = 0;
 
 		for (sg_it = superGroups.begin(); sg_it != sg_end; sg_it++) {
 			const SuperGroupPtr &superGroup = sg_it->second;
-			if (superGroup->options.appRoot == appRoot) {
+			if (appRoot == superGroup->options.appRoot) {
 				result++;
 				superGroup->restart(superGroup->options);
 			}
@@ -1437,7 +1488,13 @@ public:
 					result << "  (restarting...)" << endl;
 				}
 				if (group->spawning()) {
-					result << "  (spawning new process...)" << endl;
+					if (group->processesBeingSpawned == 0) {
+						result << "  (spawning...)" << endl;
+					} else {
+						result << "  (spawning " << group->processesBeingSpawned << " new " <<
+							maybePluralize(group->processesBeingSpawned, "process", "processes") <<
+							"...)" << endl;
+					}
 				}
 				result << "  Requests in queue: " << group->getWaitlist.size() << endl;
 				inspectProcessList(options, result, group, group->enabledProcesses);
@@ -1462,7 +1519,7 @@ public:
 		
 		result << "<process_count>" << getProcessCount(false) << "</process_count>";
 		result << "<max>" << max << "</max>";
-		result << "<utilization>" << utilization(false) << "</utilization>";
+		result << "<capacity_used>" << capacityUsed(false) << "</capacity_used>";
 		result << "<get_wait_list_size>" << getWaitlist.size() << "</get_wait_list_size>";
 
 		if (includeSecrets) {
@@ -1486,7 +1543,7 @@ public:
 			result << "<name>" << escapeForXml(superGroup->name) << "</name>";
 			result << "<state>" << superGroup->getStateName() << "</state>";
 			result << "<get_wait_list_size>" << superGroup->getWaitlist.size() << "</get_wait_list_size>";
-			result << "<utilization>" << superGroup->utilization() << "</utilization>";
+			result << "<capacity_used>" << superGroup->capacityUsed() << "</capacity_used>";
 			if (includeSecrets) {
 				result << "<secret>" << escapeForXml(superGroup->secret) << "</secret>";
 			}
