@@ -125,7 +125,7 @@ public:
  * This means that a Group outlives all its Processes, a Process outlives all
  * its Sessions, and a Process also outlives the OS process.
  */
-class Process: public enable_shared_from_this<Process> {
+class Process: public boost::enable_shared_from_this<Process> {
 private:
 	friend class Group;
 	
@@ -136,10 +136,10 @@ private:
 	 * Should never be NULL because a Group should outlive all of its Processes.
 	 * Read-only; only set once during initialization.
 	 */
-	weak_ptr<Group> group;
+	boost::weak_ptr<Group> group;
 	
 	/** A subset of 'sockets': all sockets that speak the
-	 * "session" protocol, sorted by socket.utilization(). */
+	 * "session" protocol, sorted by socket.busyness(). */
 	PriorityQueue<Socket> sessionSockets;
 	
 	/** The iterator inside the associated Group's process list. */
@@ -147,13 +147,41 @@ private:
 	/** The handle inside the associated Group's process priority queue. */
 	PriorityQueue<Process>::Handle pqHandle;
 
+	static bool
+	isZombie(pid_t pid) {
+		string filename = "/proc/" + toString(pid) + "/status";
+		FILE *f = fopen(filename.c_str(), "r");
+		if (f == NULL) {
+			// Don't know.
+			return false;
+		}
+
+		bool result = false;
+		while (!feof(f)) {
+			char buf[512];
+			const char *line;
+
+			line = fgets(buf, sizeof(buf), f);
+			if (line == NULL) {
+				break;
+			}
+			if (strcmp(line, "State:	Z (zombie)\n") == 0) {
+				// Is a zombie.
+				result = true;
+				break;
+			}
+		}
+		fclose(f);
+		return result;
+	}
+
 	void indexSessionSockets() {
 		SocketList::iterator it;
 		concurrency = 0;
 		for (it = sockets->begin(); it != sockets->end(); it++) {
 			Socket *socket = &(*it);
 			if (socket->protocol == "session" || socket->protocol == "http_session") {
-				socket->pqHandle = sessionSockets.push(socket, socket->utilization());
+				socket->pqHandle = sessionSockets.push(socket, socket->busyness());
 				if (concurrency != -1) {
 					if (socket->concurrency == 0) {
 						// If one of the sockets has a concurrency of
@@ -181,7 +209,11 @@ public:
 	SafeLibev * const libev;
 	/** Process PID. */
 	pid_t pid;
-	/** UUID for this process, randomly generated and will never appear again. */
+	/** An ID that uniquely identifies this Process in the Group, for
+	 * use in implementing sticky sessions. Set by Group::attach(). */
+	unsigned int stickySessionId;
+	/** UUID for this process, randomly generated and extremely unlikely to ever
+	 * appear again in this universe. */
 	string gupid;
 	string connectPassword;
 	/** Admin socket, see class description. */
@@ -298,6 +330,7 @@ public:
 		: pqHandle(NULL),
 		  libev(_libev.get()),
 		  pid(_pid),
+		  stickySessionId(0),
 		  gupid(_gupid),
 		  connectPassword(_connectPassword),
 		  adminSocket(_adminSocket),
@@ -317,20 +350,20 @@ public:
 	{
 		SpawnerConfigPtr config;
 		if (_config == NULL) {
-			config = make_shared<SpawnerConfig>();
+			config = boost::make_shared<SpawnerConfig>();
 		} else {
 			config = _config;
 		}
 
 		if (_adminSocket != -1) {
-			PipeWatcherPtr watcher = make_shared<PipeWatcher>(_adminSocket,
-				"stdout", pid, config->forwardStdout);
+			PipeWatcherPtr watcher = boost::make_shared<PipeWatcher>(_adminSocket,
+				"stdout", pid);
 			watcher->initialize();
 			watcher->start();
 		}
 		if (_errorPipe != -1) {
-			PipeWatcherPtr watcher = make_shared<PipeWatcher>(_errorPipe,
-				"stderr", pid, config->forwardStderr);
+			PipeWatcherPtr watcher = boost::make_shared<PipeWatcher>(_errorPipe,
+				"stderr", pid);
 			watcher->initialize();
 			watcher->start();
 		}
@@ -458,21 +491,38 @@ public:
 	 */
 	bool osProcessExists() const {
 		if (!dummy && m_osProcessExists) {
-			// Once we detect that a process is gone.
-			m_osProcessExists = syscalls::kill(pid, 0) == 0 || errno != ESRCH;
+			if (syscalls::kill(pid, 0) == 0) {
+				/* On some environments, e.g. Heroku, the init process does
+				 * not properly reap adopted zombie processes, which can interfere
+				 * with our process existance check. To work around this, we
+				 * explicitly check whether or not the process has become a zombie.
+				 */
+				m_osProcessExists = !isZombie(pid);
+			} else {
+				m_osProcessExists = errno != ESRCH;
+			}
 			return m_osProcessExists;
 		} else {
 			return false;
 		}
 	}
+
+	/** Kill the OS process with the given signal. */
+	int kill(int signo) {
+		if (osProcessExists()) {
+			return syscalls::kill(pid, signo);
+		} else {
+			return 0;
+		}
+	}
 	
-	int utilization() const {
+	int busyness() const {
 		/* Different processes within a Group may have different
 		 * 'concurrency' values. We want:
 		 * - Group.pqueue to sort the processes from least used to most used.
 		 * - to give processes with concurrency == 0 more priority over processes
 		 *   with concurrency > 0.
-		 * Therefore, we describe our utilization as a percentage of 'concurrency', with
+		 * Therefore, we describe our busyness as a percentage of 'concurrency', with
 		 * the percentage value in [0..INT_MAX] instead of [0..1].
 		 */
 		if (concurrency == 0) {
@@ -487,13 +537,21 @@ public:
 		}
 	}
 	
-	// TODO: remove this
-	bool atFullCapacity() const {
-		return atFullUtilization();
+	/**
+	 * Whether we've reached the maximum number of concurrent sessions for this
+	 * process.
+	 */
+	bool isTotallyBusy() const {
+		return concurrency != 0 && sessions >= concurrency;
 	}
 
-	bool atFullUtilization() const {
-		return concurrency != 0 && sessions >= concurrency;
+	/**
+	 * Whether a get() request can be routed to this process, assuming that
+	 * the sticky session ID (if any) matches. This is only not the case
+	 * if this process is totally busy.
+	 */
+	bool canBeRoutedTo() const {
+		return !isTotallyBusy();
 	}
 	
 	/**
@@ -507,15 +565,14 @@ public:
 	 */
 	SessionPtr newSession() {
 		Socket *socket = sessionSockets.pop();
-		if (socket->atFullCapacity()) {
+		if (socket->isTotallyBusy()) {
 			return SessionPtr();
 		} else {
 			socket->sessions++;
 			this->sessions++;
-			processed++;
-			socket->pqHandle = sessionSockets.push(socket, socket->utilization());
+			socket->pqHandle = sessionSockets.push(socket, socket->busyness());
 			lastUsed = SystemTime::getUsec();
-			return make_shared<Session>(shared_from_this(), socket);
+			return boost::make_shared<Session>(shared_from_this(), socket);
 		}
 	}
 	
@@ -527,8 +584,9 @@ public:
 		
 		socket->sessions--;
 		this->sessions--;
-		sessionSockets.decrease(socket->pqHandle, socket->utilization());
-		assert(!atFullUtilization());
+		processed++;
+		sessionSockets.decrease(socket->pqHandle, socket->busyness());
+		assert(!isTotallyBusy());
 	}
 
 	/**
@@ -543,11 +601,12 @@ public:
 	template<typename Stream>
 	void inspectXml(Stream &stream, bool includeSockets = true) const {
 		stream << "<pid>" << pid << "</pid>";
+		stream << "<sticky_session_id>" << stickySessionId << "</sticky_session_id>";
 		stream << "<gupid>" << gupid << "</gupid>";
 		stream << "<connect_password>" << connectPassword << "</connect_password>";
 		stream << "<concurrency>" << concurrency << "</concurrency>";
 		stream << "<sessions>" << sessions << "</sessions>";
-		stream << "<utilization>" << utilization() << "</utilization>";
+		stream << "<busyness>" << busyness() << "</busyness>";
 		stream << "<processed>" << processed << "</processed>";
 		stream << "<spawner_creation_time>" << spawnerCreationTime << "</spawner_creation_time>";
 		stream << "<spawn_start_time>" << spawnStartTime << "</spawn_start_time>";
