@@ -84,7 +84,7 @@ private:
 	
 	typedef MessageServer::CommonClientContext CommonClientContext;
 	
-	shared_ptr<RequestHandler> requestHandler;
+	boost::shared_ptr<RequestHandler> requestHandler;
 	PoolPtr pool;
 	
 	
@@ -122,17 +122,8 @@ private:
 			return false;
 		}
 
-		VariantMap map;
-		vector<string>::const_iterator it = args.begin(), end = args.end();
-		it++;
-		while (it != end) {
-			const string &key = *it;
-			it++;
-			const string &value = *it;
-			map.set(key, value);
-			it++;
-		}
-		writeScalarMessage(commonContext.fd, pool->inspect(Pool::InspectOptions(map)));
+		VariantMap options = argsToOptions(args);
+		writeScalarMessage(commonContext.fd, pool->inspect(Pool::InspectOptions(options)));
 		return true;
 	}
 	
@@ -155,6 +146,22 @@ private:
 		writeScalarMessage(commonContext.fd, oxt::thread::all_backtraces());
 	}
 
+	void processRestartAppGroup(CommonClientContext &commonContext, SpecificContext *specificContext,
+		const vector<string> &args)
+	{
+		TRACE_POINT();
+		commonContext.requireRights(Account::RESTART);
+		VariantMap options = argsToOptions(args, 2);
+		RestartMethod method = RM_DEFAULT;
+		if (options.get("method", false) == "blocking") {
+			method = RM_BLOCKING;
+		} else if (options.get("method", false) == "rolling") {
+			method = RM_ROLLING;
+		}
+		bool result = pool->restartGroupByName(args[1], method);
+		writeArrayMessage(commonContext.fd, result ? "true" : "false", NULL);
+	}
+
 	void processRequests(CommonClientContext &commonContext, SpecificContext *specificContext,
 		const vector<string> &args)
 	{
@@ -166,13 +173,13 @@ private:
 	}
 	
 public:
-	RemoteController(const shared_ptr<RequestHandler> &requestHandler, const PoolPtr &pool) {
+	RemoteController(const boost::shared_ptr<RequestHandler> &requestHandler, const PoolPtr &pool) {
 		this->requestHandler = requestHandler;
 		this->pool = pool;
 	}
 	
 	virtual MessageServer::ClientContextPtr newClient(CommonClientContext &commonContext) {
-		return make_shared<SpecificContext>();
+		return boost::make_shared<SpecificContext>();
 	}
 	
 	virtual bool processMessage(CommonClientContext &commonContext,
@@ -191,6 +198,8 @@ public:
 				processToXml(commonContext, specificContext, args);
 			} else if (isCommand(args, "backtraces", 0)) {
 				processBacktraces(commonContext, specificContext, args);
+			} else if (isCommand(args, "restart_app_group", 1, 99)) {
+				processRestartAppGroup(commonContext, specificContext, args);
 			} else if (isCommand(args, "requests", 0)) {
 				processRequests(commonContext, specificContext, args);
 			} else {
@@ -259,10 +268,10 @@ private:
 	AccountsDatabasePtr accountsDatabase;
 	MessageServerPtr messageServer;
 	ResourceLocator resourceLocator;
-	shared_ptr<RequestHandler> requestHandler;
-	shared_ptr<oxt::thread> prestarterThread;
-	shared_ptr<oxt::thread> messageServerThread;
-	shared_ptr<oxt::thread> eventLoopThread;
+	boost::shared_ptr<RequestHandler> requestHandler;
+	boost::shared_ptr<oxt::thread> prestarterThread;
+	boost::shared_ptr<oxt::thread> messageServerThread;
+	boost::shared_ptr<oxt::thread> eventLoopThread;
 	EventFd exitEvent;
 	
 	/**
@@ -329,8 +338,8 @@ private:
 	 */
 	void lowerPrivilege(const string &username, const string &groupname) {
 		struct passwd *userEntry;
-		struct group  *groupEntry;
-		int            e;
+		gid_t gid;
+		int e;
 		
 		userEntry = getpwnam(username.c_str());
 		if (userEntry == NULL) {
@@ -338,8 +347,8 @@ private:
 				"HelperAgent's privilege to that of user '") + username +
 				"': user does not exist.");
 		}
-		groupEntry = getgrnam(groupname.c_str());
-		if (groupEntry == NULL) {
+		gid = lookupGid(groupname);
+		if (gid == (gid_t) -1) {
 			throw NonExistentGroupException(string("Unable to lower Passenger "
 				"HelperAgent's privilege to that of user '") + username +
 				"': user does not exist.");
@@ -351,7 +360,7 @@ private:
 				"privilege to that of user '") + username +
 				"': cannot set supplementary groups for this user", e);
 		}
-		if (setgid(groupEntry->gr_gid) != 0) {
+		if (setgid(gid) != 0) {
 			e = errno;
 			throw SystemException(string("Unable to lower Passenger HelperAgent's "
 				"privilege to that of user '") + username +
@@ -421,12 +430,13 @@ public:
 		UPDATE_TRACE_POINT();
 		generation = serverInstanceDir.getGeneration(options.generationNumber);
 		startListening();
-		accountsDatabase = make_shared<AccountsDatabase>();
+		accountsDatabase = boost::make_shared<AccountsDatabase>();
 		accountsDatabase->add("_passenger-status", options.adminToolStatusPassword, false,
 			Account::INSPECT_BASIC_INFO | Account::INSPECT_SENSITIVE_INFO |
-			Account::INSPECT_BACKTRACES | Account::INSPECT_REQUESTS);
+			Account::INSPECT_BACKTRACES | Account::INSPECT_REQUESTS |
+			Account::RESTART);
 		accountsDatabase->add("_web_server", options.exitPassword, false, Account::EXIT);
-		messageServer = make_shared<MessageServer>(
+		messageServer = boost::make_shared<MessageServer>(
 			parseUnixSocketAddress(options.adminSocketAddress), accountsDatabase);
 		
 		createFile(generation->getPath() + "/helper_agent.pid",
@@ -437,7 +447,7 @@ public:
 		}
 
 		UPDATE_TRACE_POINT();
-		randomGenerator = make_shared<RandomGenerator>();
+		randomGenerator = boost::make_shared<RandomGenerator>();
 		// Check whether /dev/urandom is actually random.
 		// https://code.google.com/p/phusion-passenger/issues/detail?id=516
 		if (randomGenerator->generateByteString(16) == randomGenerator->generateByteString(16)) {
@@ -446,21 +456,20 @@ public:
 		}
 		
 		UPDATE_TRACE_POINT();
-		loggerFactory = make_shared<UnionStation::LoggerFactory>(options.loggingAgentAddress,
+		loggerFactory = boost::make_shared<UnionStation::LoggerFactory>(options.loggingAgentAddress,
 			"logging", options.loggingAgentPassword);
-		spawnerFactory = make_shared<SpawnerFactory>(poolLoop.safe,
-			resourceLocator, generation, make_shared<SpawnerConfig>(randomGenerator));
-		pool = make_shared<Pool>(poolLoop.safe.get(), spawnerFactory, loggerFactory,
-			randomGenerator);
+		spawnerFactory = boost::make_shared<SpawnerFactory>(poolLoop.safe,
+			resourceLocator, generation, boost::make_shared<SpawnerConfig>(randomGenerator));
+		pool = boost::make_shared<Pool>(spawnerFactory, loggerFactory,
+			randomGenerator, &options);
 		pool->initialize();
 		pool->setMax(options.maxPoolSize);
-		//pool->setMaxPerApp(maxInstancesPerApp);
 		pool->setMaxIdleTime(options.poolIdleTime * 1000000);
 		
-		requestHandler = make_shared<RequestHandler>(requestLoop.safe,
+		requestHandler = boost::make_shared<RequestHandler>(requestLoop.safe,
 			requestSocket, pool, options);
 
-		messageServer->addHandler(make_shared<RemoteController>(requestHandler, pool));
+		messageServer->addHandler(boost::make_shared<RemoteController>(requestHandler, pool));
 		messageServer->addHandler(ptr(new ExitHandler(exitEvent)));
 
 		sigquitWatcher.set(requestLoop.loop);
@@ -475,7 +484,7 @@ public:
 			messageServer->getSocketFilename().c_str(),
 			NULL);
 		
-		function<void ()> func = boost::bind(prestartWebApps,
+		boost::function<void ()> func = boost::bind(prestartWebApps,
 			resourceLocator,
 			options.defaultRubyCommand,
 			options.prestartUrls
@@ -506,7 +515,24 @@ public:
 		requestHandler.reset();
 
 		if (!options.requestSocketLink.empty()) {
-			syscalls::unlink(options.requestSocketLink.c_str());
+			char path[PATH_MAX + 1];
+			ssize_t ret;
+			bool shouldUnlink;
+
+			ret = readlink(options.requestSocketLink.c_str(), path, PATH_MAX);
+			if (ret != -1) {
+				path[ret] = '\0';
+				// Only unlink if a new Flying Passenger instance hasn't overwritten the
+				// symlink.
+				// https://code.google.com/p/phusion-passenger/issues/detail?id=939
+				shouldUnlink = getRequestSocketFilename() == path;
+			} else {
+				shouldUnlink = true;
+			}
+
+			if (shouldUnlink) {
+				syscalls::unlink(options.requestSocketLink.c_str());
+			}
 		}
 		
 		P_TRACE(2, "All threads have been shut down.");
@@ -514,7 +540,7 @@ public:
 	
 	void mainLoop() {
 		TRACE_POINT();
-		function<void ()> func;
+		boost::function<void ()> func;
 
 		func = boost::bind(&MessageServer::mainLoop, messageServer.get());
 		messageServerThread = ptr(new oxt::thread(
@@ -563,6 +589,7 @@ public:
 			 */
 			P_DEBUG("Received command to exit gracefully. "
 				"Waiting until 5 seconds after all clients have disconnected...");
+			requestHandler->resetInactivityTime();
 			while (requestHandler->inactivityTime() < 5000) {
 				syscalls::usleep(250000);
 			}
@@ -586,13 +613,25 @@ public:
 int
 main(int argc, char *argv[]) {
 	TRACE_POINT();
-	AgentOptions options(initializeAgent(argc, argv, "PassengerHelperAgent"));
+	AgentOptionsPtr options;
+	try {
+		options = boost::make_shared<AgentOptions>(
+			initializeAgent(argc, argv, "PassengerHelperAgent"));
+	} catch (const VariantMap::MissingKeyException &e) {
+		fprintf(stderr, "Option required: %s\n", e.getKey().c_str());
+		return 1;
+	}
+	if (options->testBinary) {
+		printf("PASS\n");
+		exit(0);
+	}
+
 	P_DEBUG("Starting PassengerHelperAgent...");
 	MultiLibeio::init();
 	
 	try {
 		UPDATE_TRACE_POINT();
-		Server server(FileDescriptor(FEEDBACK_FD), options);
+		Server server(FileDescriptor(FEEDBACK_FD), *options);
 		P_WARN("PassengerHelperAgent online, listening at unix:" <<
 			server.getRequestSocketFilename());
 		
