@@ -90,6 +90,7 @@
 #include <sys/types.h>
 #include <arpa/inet.h>
 #include <sys/un.h>
+#include <utility>
 #include <typeinfo>
 #include <cassert>
 #include <cctype>
@@ -120,12 +121,21 @@ using namespace ApplicationPool2;
 
 class RequestHandler;
 
+#define MAX_STATUS_HEADER_SIZE 64
+
+#define RH_ERROR(client, x) P_ERROR("[Client " << client->name() << "] " << x)
 #define RH_WARN(client, x) P_WARN("[Client " << client->name() << "] " << x)
 #define RH_DEBUG(client, x) P_DEBUG("[Client " << client->name() << "] " << x)
 #define RH_TRACE(client, level, x) P_TRACE(level, "[Client " << client->name() << "] " << x)
 
+#define RH_LOG_EVENT(client, eventName) \
+	char _clientName[7 + 8]; \
+	snprintf(_clientName, sizeof(_clientName), "Client %d", client->fdnum); \
+	TRACE_POINT_WITH_DATA(_clientName); \
+	RH_TRACE(client, 3, "Event: " eventName)
 
-class Client: public enable_shared_from_this<Client> {
+
+class Client: public boost::enable_shared_from_this<Client> {
 private:
 	struct ev_loop *getLoop() const;
 	const SafeLibevPtr &getSafeLibev() const;
@@ -152,6 +162,7 @@ private:
 
 	static size_t onAppInputData(const EventedBufferedInputPtr &source, const StaticString &data);
 	static void onAppInputChunk(const char *data, size_t size, void *userData);
+	static void onAppInputChunkEnd(void *userData);
 	static void onAppInputError(const EventedBufferedInputPtr &source, const char *message, int errnoCode);
 	
 	void onAppOutputWritable(ev::io &io, int revents);
@@ -174,6 +185,7 @@ private:
 		contentLength = 0;
 		clientBodyAlreadyRead = 0;
 		checkoutSessionAfterCommit = false;
+		stickySession = false;
 		sessionCheckedOut = false;
 		sessionCheckoutTry = 0;
 		responseHeaderSeen = false;
@@ -273,6 +285,7 @@ public:
 	bool requestBodyIsBuffered;
 	bool sessionCheckedOut;
 	bool checkoutSessionAfterCommit;
+	bool stickySession;
 
 	bool responseHeaderSeen;
 	bool chunkedResponse;
@@ -283,19 +296,19 @@ public:
 	Client() {
 		fdnum = -1;
 
-		clientInput = make_shared< EventedBufferedInput<> >();
+		clientInput = boost::make_shared< EventedBufferedInput<> >();
 		clientInput->onData   = onClientInputData;
 		clientInput->onError  = onClientInputError;
 		clientInput->userData = this;
 		
-		clientBodyBuffer = make_shared<FileBackedPipe>("/tmp");
+		clientBodyBuffer = boost::make_shared<FileBackedPipe>("/tmp");
 		clientBodyBuffer->userData  = this;
 		clientBodyBuffer->onData    = onClientBodyBufferData;
 		clientBodyBuffer->onEnd     = onClientBodyBufferEnd;
 		clientBodyBuffer->onError   = onClientBodyBufferError;
 		clientBodyBuffer->onCommit  = onClientBodyBufferCommit;
 
-		clientOutputPipe = make_shared<FileBackedPipe>("/tmp");
+		clientOutputPipe = boost::make_shared<FileBackedPipe>("/tmp");
 		clientOutputPipe->userData  = this;
 		clientOutputPipe->onData    = onClientOutputPipeData;
 		clientOutputPipe->onEnd     = onClientOutputPipeEnd;
@@ -305,7 +318,7 @@ public:
 		clientOutputWatcher.set<Client, &Client::onClientOutputWritable>(this);
 
 		
-		appInput = make_shared< EventedBufferedInput<> >();
+		appInput = boost::make_shared< EventedBufferedInput<> >();
 		appInput->onData   = onAppInputData;
 		appInput->onError  = onAppInputError;
 		appInput->userData = this;
@@ -317,6 +330,7 @@ public:
 
 
 		responseDechunker.onData = onAppInputChunk;
+		responseDechunker.onEnd = onAppInputChunkEnd;
 		responseDechunker.userData = this;
 		
 
@@ -457,6 +471,8 @@ public:
 	}
 
 	bool shouldHalfCloseWrite() const {
+		// Many broken HTTP servers consider a half close to be a full close, so don't
+		// half close HTTP sessions.
 		return session->getProtocol() == "session";
 	}
 
@@ -534,7 +550,7 @@ public:
 	}
 };
 
-typedef shared_ptr<Client> ClientPtr;
+typedef boost::shared_ptr<Client> ClientPtr;
 
 
 class RequestHandler {
@@ -623,6 +639,10 @@ private:
 	// GDB helper function, implemented in .cpp file to prevent inlining.
 	Client *getClientPointer(const ClientPtr &client);
 
+	void doResetInactivityTime() {
+		inactivityTimer.reset();
+	}
+
 	void getInactivityTime(unsigned long long *result) const {
 		*result = inactivityTimer.elapsed();
 	}
@@ -651,28 +671,38 @@ private:
 		}
 	}
 
-	void writeSimpleResponse(const ClientPtr &client, const StaticString &data) {
-		char header[256];
+	void writeSimpleResponse(const ClientPtr &client, const StaticString &data, int code = 200) {
+		char header[256], statusBuffer[50];
 		char *pos = header;
 		const char *end = header + sizeof(header) - 1;
+		const char *status;
+
+		status = getStatusCodeAndReasonPhrase(code);
+		if (status == NULL) {
+			snprintf(statusBuffer, sizeof(statusBuffer), "%d Unknown Reason-Phrase", code);
+			status = statusBuffer;
+		}
 
 		if (getBoolOption(client, "PASSENGER_STATUS_LINE", true)) {
-			pos = appendData(pos, end, "HTTP/1.1 200 OK\r\n");
+			pos += snprintf(pos, end - pos, "HTTP/1.1 %s\r\n",
+				status);
 		}
 		pos += snprintf(pos, end - pos,
-			"Status: 200 OK\r\n"
+			"Status: %s\r\n"
 			"Content-Length: %lu\r\n"
 			"Content-Type: text/html; charset=UTF-8\r\n"
 			"Cache-Control: no-cache, no-store, must-revalidate\r\n"
 			"\r\n",
-			(unsigned long) data.size());
+			status, (unsigned long) data.size());
 
 		client->clientOutputPipe->write(header, pos - header);
 		client->clientOutputPipe->write(data.data(), data.size());
 		client->clientOutputPipe->end();
 
 		if (client->useUnionStation()) {
-			client->logMessage("Status: 200 OK");
+			snprintf(header, end - header, "Status: %d %s",
+				code, status);
+			client->logMessage(header);
 		}
 	}
 
@@ -880,15 +910,22 @@ private:
 		}
 		if (begin != string::npos && end != string::npos) {
 			StaticString statusValue(headerData.data() + begin + 1, end - begin);
-			char header[statusValue.size() + 20];
-			char *pos = header;
-			const char *end = header + statusValue.size() + 20;
+			if (statusValue.size() <= MAX_STATUS_HEADER_SIZE) {
+				char header[MAX_STATUS_HEADER_SIZE + sizeof("Status: \r\n")];
+				char *pos = header;
+				const char *end = header + sizeof(header);
 
-			pos = appendData(pos, end, "Status: ");
-			pos = appendData(pos, end, statusValue);
-			pos = appendData(pos, end, "\r\n");
-			headerData.append(StaticString(header, pos - header));
-			return true;
+				pos = appendData(pos, end, "Status: ");
+				pos = appendData(pos, end, statusValue);
+				pos = appendData(pos, end, "\r\n");
+				headerData.append(StaticString(header, pos - header));
+				return true;
+			} else {
+				disconnectWithError(client, "application sent malformed response: the Status header's (" +
+					statusValue + ") exceeds the allowed limit of " +
+					toString(MAX_STATUS_HEADER_SIZE) + " bytes.");
+				return false;
+			}
 		} else {
 			disconnectWithError(client, "application sent malformed response: the HTTP status line is invalid.");
 			return false;
@@ -1014,6 +1051,16 @@ private:
 			headerData.append("X-Powered-By: Phusion Passenger\r\n");
 		}
 
+		// Add sticky session ID.
+		if (client->stickySession && client->session != NULL) {
+			StaticString cookieName = getStickySessionCookieName(client);
+			headerData.append("Set-Cookie: ");
+			headerData.append(cookieName.data(), cookieName.size());
+			headerData.append("=");
+			headerData.append(toString(client->session->getStickySessionId()));
+			headerData.append("; HttpOnly\r\n");
+		}
+
 		// Add Date header. https://code.google.com/p/phusion-passenger/issues/detail?id=485
 		if (lookupHeader(headerData, "Date", "date").empty()) {
 			char dateStr[60];
@@ -1023,8 +1070,8 @@ private:
 			struct tm the_tm;
 
 			pos = appendData(pos, end, "Date: ");
-			localtime_r(&the_time, &the_tm);
-			pos += strftime(pos, end - pos, "%a, %d %b %G %H:%M:%S %Z", &the_tm);
+			gmtime_r(&the_time, &the_tm);
+			pos += strftime(pos, end - pos, "%a, %d %b %Y %H:%M:%S %Z", &the_tm);
 			pos = appendData(pos, end, "\r\n");
 			headerData.append(dateStr, pos - dateStr);
 		}
@@ -1058,6 +1105,7 @@ private:
 	}
 
 	size_t onAppInputData(const ClientPtr &client, const StaticString &data) {
+		RH_LOG_EVENT(client, "onAppInputData");
 		if (!client->connected()) {
 			return 0;
 		}
@@ -1099,20 +1147,31 @@ private:
 	}
 
 	void onAppInputChunk(const ClientPtr &client, const StaticString &data) {
+		RH_LOG_EVENT(client, "onAppInputChunk");
 		writeToClientOutputPipe(client, data);
 	}
 
+	void onAppInputChunkEnd(const ClientPtr &client) {
+		RH_LOG_EVENT(client, "onAppInputChunkEnd");
+		onAppInputEof(client);
+	}
+
 	void onAppInputEof(const ClientPtr &client) {
-		if (!client->connected()) {
+		RH_LOG_EVENT(client, "onAppInputEof");
+		// Check for session == NULL in order to avoid executing the code twice on
+		// responses with chunked encoding.
+		if (!client->connected() || client->session == NULL) {
 			return;
 		}
 
 		RH_DEBUG(client, "Application sent EOF");
+		client->session.reset();
 		client->endScopeLog(&client->scopeLogs.requestProxying);
 		client->clientOutputPipe->end();
 	}
 
 	void onAppInputError(const ClientPtr &client, const char *message, int errorCode) {
+		RH_LOG_EVENT(client, "onAppInputError");
 		if (!client->connected()) {
 			return;
 		}
@@ -1132,6 +1191,7 @@ private:
 	}
 
 	void onClientOutputPipeCommit(const ClientPtr &client) {
+		RH_LOG_EVENT(client, "onClientOutputPipeCommit");
 		if (!client->connected()) {
 			return;
 		}
@@ -1155,6 +1215,7 @@ private:
 	void onClientOutputPipeData(const ClientPtr &client, const char *data,
 		size_t size, const FileBackedPipe::ConsumeCallback &consumed)
 	{
+		RH_LOG_EVENT(client, "onClientOutputPipeData");
 		if (!client->connected()) {
 			return;
 		}
@@ -1185,6 +1246,7 @@ private:
 	}
 
 	void onClientOutputPipeEnd(const ClientPtr &client) {
+		RH_LOG_EVENT(client, "onClientOutputPipeEnd");
 		if (!client->connected()) {
 			return;
 		}
@@ -1195,6 +1257,7 @@ private:
 	}
 
 	void onClientOutputPipeError(const ClientPtr &client, int errorCode) {
+		RH_LOG_EVENT(client, "onClientOutputPipeError");
 		if (!client->connected()) {
 			return;
 		}
@@ -1207,6 +1270,7 @@ private:
 	}
 
 	void onClientOutputWritable(const ClientPtr &client) {
+		RH_LOG_EVENT(client, "onClientOutputWritable");
 		if (!client->connected()) {
 			return;
 		}
@@ -1237,7 +1301,8 @@ private:
 		if (accept4Available) {
 			FileDescriptor fd(callAccept4(requestSocket,
 				(struct sockaddr *) &u, &addrlen, O_NONBLOCK));
-			if (fd == -1 && errno == ENOSYS) {
+			// FreeBSD returns EINVAL if accept4() is called with invalid flags.
+			if (fd == -1 && (errno == ENOSYS || errno == EINVAL)) {
 				accept4Available = false;
 				return acceptNonBlockingSocket(sock);
 			} else {
@@ -1292,9 +1357,9 @@ private:
 					"\r\n"
 					"Benchmark point: after_accept\n");
 			} else {
-				ClientPtr client = make_shared<Client>();
+				ClientPtr client = boost::make_shared<Client>();
 				client->associate(this, fd);
-				clients.insert(make_pair<int, ClientPtr>(fd, client));
+				clients.insert(make_pair((int) fd, client));
 				acceptedClients[count] = client;
 				count++;
 				RH_DEBUG(client, "New client accepted; new client count = " << clients.size());
@@ -1312,7 +1377,7 @@ private:
 
 
 	size_t onClientInputData(const ClientPtr &client, const StaticString &data) {
-		RH_TRACE(client, 3, "Event: onClientInputData");
+		RH_LOG_EVENT(client, "onClientInputData");
 		if (!client->connected()) {
 			return 0;
 		}
@@ -1363,7 +1428,7 @@ private:
 	}
 
 	void onClientEof(const ClientPtr &client) {
-		RH_TRACE(client, 3, "Event: onClientEof; client sent EOF");
+		RH_LOG_EVENT(client, "onClientEof; client sent EOF");
 		switch (client->state) {
 		case Client::BUFFERING_REQUEST_BODY:
 			state_bufferingRequestBody_onClientEof(client);
@@ -1378,7 +1443,7 @@ private:
 	}
 
 	void onClientInputError(const ClientPtr &client, const char *message, int errnoCode) {
-		RH_TRACE(client, 3, "Event: onClientInputError");
+		RH_LOG_EVENT(client, "onClientInputError");
 		if (!client->connected()) {
 			return;
 		}
@@ -1399,7 +1464,7 @@ private:
 
 
 	void onClientBodyBufferData(const ClientPtr &client, const char *data, size_t size, const FileBackedPipe::ConsumeCallback &consumed) {
-		RH_TRACE(client, 3, "Event: onClientBodyBufferData");
+		RH_LOG_EVENT(client, "onClientBodyBufferData");
 		if (!client->connected()) {
 			return;
 		}
@@ -1414,7 +1479,7 @@ private:
 	}
 
 	void onClientBodyBufferError(const ClientPtr &client, int errorCode) {
-		RH_TRACE(client, 3, "Event: onClientBodyBufferError");
+		RH_LOG_EVENT(client, "onClientBodyBufferError");
 		if (!client->connected()) {
 			return;
 		}
@@ -1427,7 +1492,7 @@ private:
 	}
 
 	void onClientBodyBufferEnd(const ClientPtr &client) {
-		RH_TRACE(client, 3, "Event: onClientBodyBufferEnd");
+		RH_LOG_EVENT(client, "onClientBodyBufferEnd");
 		if (!client->connected()) {
 			return;
 		}
@@ -1442,7 +1507,7 @@ private:
 	}
 
 	void onClientBodyBufferCommit(const ClientPtr &client) {
-		RH_TRACE(client, 3, "Event: onClientBodyBufferCommit");
+		RH_LOG_EVENT(client, "onClientBodyBufferCommit");
 		if (!client->connected()) {
 			return;
 		}
@@ -1457,7 +1522,7 @@ private:
 	}
 
 	void onAppOutputWritable(const ClientPtr &client) {
-		RH_TRACE(client, 3, "Event: onAppOutputWritable");
+		RH_LOG_EVENT(client, "onAppOutputWritable");
 		if (!client->connected()) {
 			return;
 		}
@@ -1476,7 +1541,7 @@ private:
 
 
 	void onTimeout(const ClientPtr &client) {
-		RH_TRACE(client, 3, "Event: onTimeout");
+		RH_LOG_EVENT(client, "onTimeout");
 		if (!client->connected()) {
 			return;
 		}
@@ -1600,6 +1665,13 @@ private:
 		}
 	}
 
+	static void fillPoolOption(const ClientPtr &client, unsigned int &field, const StaticString &name) {
+		ScgiRequestParser::const_iterator it = client->scgiParser.getHeaderIterator(name);
+		if (it != client->scgiParser.end()) {
+			field = stringToUint(it->second);
+		}
+	}
+
 	static void fillPoolOption(const ClientPtr &client, unsigned long &field, const StaticString &name) {
 		ScgiRequestParser::const_iterator it = client->scgiParser.getHeaderIterator(name);
 		if (it != client->scgiParser.end()) {
@@ -1614,6 +1686,13 @@ private:
 		}
 	}
 
+	static void fillPoolOptionSecToMsec(const ClientPtr &client, unsigned int &field, const StaticString &name) {
+		ScgiRequestParser::const_iterator it = client->scgiParser.getHeaderIterator(name);
+		if (it != client->scgiParser.end()) {
+			field = stringToUint(it->second) * 1000;
+		}
+	}
+
 	void fillPoolOptions(const ClientPtr &client) {
 		Options &options = client->options;
 		ScgiRequestParser &parser = client->scgiParser;
@@ -1654,21 +1733,28 @@ private:
 		options.defaultGroup = this->options.defaultGroup;
 		fillPoolOption(client, options.appGroupName, "PASSENGER_APP_GROUP_NAME");
 		fillPoolOption(client, options.appType, "PASSENGER_APP_TYPE");
-		fillPoolOption(client, options.environment, "PASSENGER_ENV");
+		fillPoolOption(client, options.environment, "PASSENGER_APP_ENV");
 		fillPoolOption(client, options.ruby, "PASSENGER_RUBY");
 		fillPoolOption(client, options.python, "PASSENGER_PYTHON");
+		fillPoolOption(client, options.nodejs, "PASSENGER_NODEJS");
 		fillPoolOption(client, options.user, "PASSENGER_USER");
 		fillPoolOption(client, options.group, "PASSENGER_GROUP");
-		fillPoolOption(client, options.minProcesses, "PASSENGER_MIN_INSTANCES");
+		fillPoolOption(client, options.minProcesses, "PASSENGER_MIN_PROCESSES");
+		fillPoolOption(client, options.maxProcesses, "PASSENGER_MAX_PROCESSES");
 		fillPoolOption(client, options.maxRequests, "PASSENGER_MAX_REQUESTS");
 		fillPoolOption(client, options.spawnMethod, "PASSENGER_SPAWN_METHOD");
 		fillPoolOption(client, options.startCommand, "PASSENGER_START_COMMAND");
+		fillPoolOptionSecToMsec(client, options.startTimeout, "PASSENGER_START_TIMEOUT");
 		fillPoolOption(client, options.maxPreloaderIdleTime, "PASSENGER_MAX_PRELOADER_IDLE_TIME");
+		fillPoolOption(client, options.maxRequestQueueSize, "PASSENGER_MAX_REQUEST_QUEUE_SIZE");
 		fillPoolOption(client, options.statThrottleRate, "PASSENGER_STAT_THROTTLE_RATE");
 		fillPoolOption(client, options.restartDir, "PASSENGER_RESTART_DIR");
+		fillPoolOption(client, options.startupFile, "PASSENGER_STARTUP_FILE");
 		fillPoolOption(client, options.loadShellEnvvars, "PASSENGER_LOAD_SHELL_ENVVARS");
 		fillPoolOption(client, options.debugger, "PASSENGER_DEBUGGER");
 		fillPoolOption(client, options.raiseInternalError, "PASSENGER_RAISE_INTERNAL_ERROR");
+		setStickySessionId(client);
+		/******************/
 		
 		for (it = client->scgiParser.begin(); it != end; it++) {
 			if (!startsWith(it->first, "PASSENGER_")
@@ -1695,13 +1781,18 @@ private:
 				return;
 			}
 
-			client->options.analytics = true;
-			client->options.unionStationKey = key;
 			client->options.logger = loggerFactory->newTransaction(
 				options.getAppGroupName(), "requests", key, filters);
+			if (!client->options.logger->isNull()) {
+				client->options.analytics = true;
+				client->options.unionStationKey = key;
+			}
 			
 			client->beginScopeLog(&client->scopeLogs.requestProcessing, "request processing");
 
+			StaticString staticRequestMethod = parser.getHeader("REQUEST_METHOD");
+			client->logMessage("Request method: " + staticRequestMethod);
+
 			StaticString staticRequestURI = parser.getHeader("REQUEST_URI");
 			if (!staticRequestURI.empty()) {
 				client->logMessage("URI: " + staticRequestURI);
@@ -1718,6 +1809,53 @@ private:
 		}
 	}
 
+	void setStickySessionId(const ClientPtr &client) {
+		ScgiRequestParser &parser = client->scgiParser;
+		if (parser.getHeader("PASSENGER_STICKY_SESSION") == "true") {
+			// TODO: This is not entirely correct. Clients MAY send multiple Cookie
+			// headers, although this is in practice extremely rare.
+			// http://stackoverflow.com/questions/16305814/are-multiple-cookie-headers-allowed-in-an-http-request
+			StaticString cookie = parser.getHeader("HTTP_COOKIE");
+			StaticString cookieName = getStickySessionCookieName(client);
+			vector<StaticString> parts;
+
+			client->stickySession = true;
+			split(cookie, ';', parts);
+			foreach (StaticString part, parts) {
+				const char *begin = part.data();
+				const char *end = part.data() + part.size();
+				const char *sep;
+
+				// Skip leading whitespace in the name.
+				while (begin < end && *begin == ' ') {
+					begin++;
+				}
+				part = StaticString(begin, end - begin);
+
+				// Find the separator ('=').
+				sep = (const char *) memchr(begin, '=', end - begin);
+				if (sep != NULL) {
+					StaticString name(begin, sep - begin);
+					if (name == cookieName) {
+						// This cookie matches the one we're looking for.
+						StaticString value(sep + 1, end - (sep + 1));
+						client->options.stickySessionId = stringToUint(value);
+						return;
+					}
+				}
+			}
+		}
+	}
+
+	StaticString getStickySessionCookieName(const ClientPtr &client) const {
+		StaticString value = client->scgiParser.getHeader("PASSENGER_STICKY_SESSION_COOKIE_NAME");
+		if (value.empty()) {
+			return StaticString("_passenger_route", sizeof("_passenger_route") - 1);
+		} else {
+			return value;
+		}
+	}
+
 	size_t state_readingHeader_onClientData(const ClientPtr &client, const char *data, size_t size) {
 		ScgiRequestParser &parser = client->scgiParser;
 		size_t consumed = parser.feed(data, size);
@@ -1882,47 +2020,21 @@ private:
 
 		if (e != NULL) {
 			client->endScopeLog(&client->scopeLogs.getFromPool, false);
-			shared_ptr<SpawnException> e2 = dynamic_pointer_cast<SpawnException>(e);
-			if (e2 != NULL) {
-				if (strip(e2->getErrorPage()).empty()) {
-					RH_WARN(client, "Cannot checkout session. " << e2->what());
-					writeErrorResponse(client, e2->what());
-				} else {
-					RH_WARN(client, "Cannot checkout session. " << e2->what() <<
-						"\nError page:\n" << e2->getErrorPage());
-					writeErrorResponse(client, e2->getErrorPage(), e2.get());
+			{
+				boost::shared_ptr<RequestQueueFullException> e2 = dynamic_pointer_cast<RequestQueueFullException>(e);
+				if (e2 != NULL) {
+					writeRequestQueueFullExceptionErrorResponse(client);
+					return;
 				}
-			} else {
-				string typeName;
-				#ifdef CXX_ABI_API_AVAILABLE
-					int status;
-					char *tmp = abi::__cxa_demangle(typeid(*e).name(), 0, 0, &status);
-					if (tmp != NULL) {
-						typeName = tmp;
-						free(tmp);
-					} else {
-						typeName = typeid(*e).name();
-					}
-				#else
-					typeName = typeid(*e).name();
-				#endif
-
-				RH_WARN(client, "Cannot checkout session (exception type " <<
-					typeName << "): " << e->what());
-				
-				string response = "An internal error occurred while trying to spawn the application.\n";
-				response.append("Exception type: ");
-				response.append(typeName);
-				response.append("\nError message: ");
-				response.append(e->what());
-				shared_ptr<tracable_exception> e3 = dynamic_pointer_cast<tracable_exception>(e);
-				if (e3 != NULL) {
-					response.append("\nBacktrace:\n");
-					response.append(e3->backtrace());
-				}
-
-				writeErrorResponse(client, response);
 			}
+			{
+				boost::shared_ptr<SpawnException> e2 = dynamic_pointer_cast<SpawnException>(e);
+				if (e2 != NULL) {
+					writeSpawnExceptionErrorResponse(client, e2);
+					return;
+				}
+			}
+			writeOtherExceptionErrorResponse(client, e);
 		} else {
 			RH_DEBUG(client, "Session checked out: pid=" << session->getPid() <<
 				", gupid=" << session->getGupid());
@@ -1931,6 +2043,62 @@ private:
 		}
 	}
 
+	void writeRequestQueueFullExceptionErrorResponse(const ClientPtr &client) {
+		StaticString value = client->scgiParser.getHeader("PASSENGER_REQUEST_QUEUE_OVERFLOW_STATUS_CODE");
+		int requestQueueOverflowStatusCode = 503;
+		if (!value.empty()) {
+			requestQueueOverflowStatusCode = atoi(value.data());
+		}
+		writeSimpleResponse(client,
+			"<h1>This website is under heavy load</h1>"
+			"<p>We're sorry, too many people are accessing this website at the same "
+			"time. We're working on this problem. Please try again later.</p>",
+			requestQueueOverflowStatusCode);
+	}
+
+	void writeSpawnExceptionErrorResponse(const ClientPtr &client, const boost::shared_ptr<SpawnException> &e) {
+		if (strip(e->getErrorPage()).empty()) {
+			RH_WARN(client, "Cannot checkout session. " << e->what());
+			writeErrorResponse(client, e->what());
+		} else {
+			RH_WARN(client, "Cannot checkout session.\nError page:\n" <<
+				e->getErrorPage());
+			writeErrorResponse(client, e->getErrorPage(), e.get());
+		}
+	}
+
+	void writeOtherExceptionErrorResponse(const ClientPtr &client, const ExceptionPtr &e) {
+		string typeName;
+		#ifdef CXX_ABI_API_AVAILABLE
+			int status;
+			char *tmp = abi::__cxa_demangle(typeid(*e).name(), 0, 0, &status);
+			if (tmp != NULL) {
+				typeName = tmp;
+				free(tmp);
+			} else {
+				typeName = typeid(*e).name();
+			}
+		#else
+			typeName = typeid(*e).name();
+		#endif
+
+		RH_WARN(client, "Cannot checkout session (exception type " <<
+			typeName << "): " << e->what());
+		
+		string response = "An internal error occurred while trying to spawn the application.\n";
+		response.append("Exception type: ");
+		response.append(typeName);
+		response.append("\nError message: ");
+		response.append(e->what());
+		boost::shared_ptr<tracable_exception> e3 = dynamic_pointer_cast<tracable_exception>(e);
+		if (e3 != NULL) {
+			response.append("\nBacktrace:\n");
+			response.append(e3->backtrace());
+		}
+
+		writeErrorResponse(client, response);
+	}
+
 	void initiateSession(const ClientPtr &client) {
 		assert(client->state == Client::CHECKING_OUT_SESSION);
 		client->sessionCheckoutTry++;
@@ -1987,7 +2155,10 @@ private:
 
 		RH_TRACE(client, 2, "Sending headers to application");
 
-		if (client->session->getProtocol() == "session") {
+		if (client->session == NULL) {
+			disconnectWithError(client,
+				"Application sent EOF before we were able to send headers to it");
+		} else if (client->session->getProtocol() == "session") {
 			char sizeField[sizeof(uint32_t)];
 			SmallVector<StaticString, 10> data;
 
@@ -2032,7 +2203,7 @@ private:
 			data.append("Connection: close\r\n");
 
 			for (it = parser.begin(); it != end; it++) {
-				if (startsWith(it->first, "HTTP_")) {
+				if (startsWith(it->first, "HTTP_") && it->first != "HTTP_CONNECTION") {
 					string subheader = it->first.substr(sizeof("HTTP_") - 1);
 					string::size_type i;
 					for (i = 0; i < subheader.size(); i++) {
@@ -2090,22 +2261,27 @@ private:
 	void state_sendingHeaderToApp_onAppOutputWritable(const ClientPtr &client) {
 		state_sendingHeaderToApp_verifyInvariants(client);
 
-		ssize_t ret = gatheredWrite(client->session->fd(), NULL, 0, client->appOutputBuffer);
-		if (ret == -1) {
-			if (errno != EAGAIN && errno != EPIPE && errno != ECONNRESET) {
-				disconnectWithAppSocketWriteError(client, errno);
+		if (client->session == NULL) {
+			disconnectWithError(client,
+				"Application sent EOF before we were able to send headers to it");
+		} else {
+			ssize_t ret = gatheredWrite(client->session->fd(), NULL, 0, client->appOutputBuffer);
+			if (ret == -1) {
+				if (errno != EAGAIN && errno != EPIPE && errno != ECONNRESET) {
+					disconnectWithAppSocketWriteError(client, errno);
+				}
+				// TODO: what about other errors?
+			} else if (client->appOutputBuffer.empty()) {
+				client->appOutputWatcher.stop();
+				sendBodyToApp(client);
 			}
-			// TODO: what about other errors?
-		} else if (client->appOutputBuffer.empty()) {
-			client->appOutputWatcher.stop();
-			sendBodyToApp(client);
 		}
 	}
 
 
 	/******* State: FORWARDING_BODY_TO_APP *******/
 
-	void state_forwardingBodyToApp_verifyInvariants(const ClientPtr &client) {
+	void state_forwardingBodyToApp_verifyInvariants(const ClientPtr &client) const {
 		assert(client->state == Client::FORWARDING_BODY_TO_APP);
 	}
 
@@ -2142,6 +2318,14 @@ private:
 		}
 
 		RH_TRACE(client, 3, "Forwarding " << size << " bytes of client body data to application.");
+
+		if (client->session == NULL) {
+			RH_TRACE(client, 2, "Application had already sent EOF. Stop reading client input.");
+			client->clientInput->stop();
+			syscalls::shutdown(client->fd, SHUT_RD);
+			return 0;
+		}
+
 		ssize_t ret = syscalls::write(client->session->fd(), data, size);
 		int e = errno;
 		if (ret == -1) {
@@ -2179,7 +2363,7 @@ private:
 
 		RH_TRACE(client, 2, "End of (unbuffered) client body reached; done sending data to application");
 		client->clientInput->stop();
-		if (client->shouldHalfCloseWrite()) {
+		if (client->session != NULL && client->shouldHalfCloseWrite()) {
 			syscalls::shutdown(client->session->fd(), SHUT_WR);
 		}
 	}
@@ -2206,6 +2390,14 @@ private:
 		assert(client->requestBodyIsBuffered);
 
 		RH_TRACE(client, 3, "Forwarding " << size << " bytes of buffered client body data to application.");
+
+		if (client->session == NULL) {
+			RH_TRACE(client, 2, "Application had already sent EOF. Stop reading client input.");
+			syscalls::shutdown(client->fd, SHUT_RD);
+			consumed(0, true);
+			return;
+		}
+
 		ssize_t ret = syscalls::write(client->session->fd(), data, size);
 		if (ret == -1) {
 			int e = errno;
@@ -2232,7 +2424,7 @@ private:
 		assert(client->requestBodyIsBuffered);
 
 		RH_TRACE(client, 2, "End of (buffered) client body reached; done sending data to application");
-		if (client->shouldHalfCloseWrite()) {
+		if (client->session != NULL && client->shouldHalfCloseWrite()) {
 			syscalls::shutdown(client->session->fd(), SHUT_WR);
 		}
 	}
@@ -2280,6 +2472,10 @@ public:
 		}
 	}
 
+	void resetInactivityTime() {
+		libev->run(boost::bind(&RequestHandler::doResetInactivityTime, this));
+	}
+
 	unsigned long long inactivityTime() const {
 		unsigned long long result;
 		libev->run(boost::bind(&RequestHandler::getInactivityTime, this, &result));
