@@ -30,6 +30,8 @@
 #include <boost/foreach.hpp>
 #include <boost/enable_shared_from_this.hpp>
 #include <string>
+#include <utility>
+#include <vector>
 
 #include <sys/select.h>
 #include <sys/types.h>
@@ -49,11 +51,13 @@
 #include <Logging.h>
 #include <Exceptions.h>
 #include <StaticString.h>
+#include <Hooks.h>
 #include <ResourceLocator.h>
 #include <Utils.h>
 #include <Utils/Base64.h>
 #include <Utils/Timer.h>
 #include <Utils/ScopeGuard.h>
+#include <Utils/StrIntUtils.h>
 #include <Utils/IOUtils.h>
 #include <Utils/MessageIO.h>
 #include <Utils/VariantMap.h>
@@ -97,6 +101,7 @@ struct WorkingObjects {
 	ServerInstanceDir::GenerationPtr generation;
 	uid_t defaultUid;
 	gid_t defaultGid;
+	vector<string> cleanupPidfiles;
 	string loggingAgentAddress;
 	string loggingAgentPassword;
 	string loggingAgentAdminAddress;
@@ -104,7 +109,7 @@ struct WorkingObjects {
 	string adminToolManipulationPassword;
 };
 
-typedef shared_ptr<WorkingObjects> WorkingObjectsPtr;
+typedef boost::shared_ptr<WorkingObjects> WorkingObjectsPtr;
 
 static string oldOomScore;
 
@@ -167,7 +172,10 @@ setOomScore(const StaticString &score) {
 	
 	f = openOomAdjFile("w", type);
 	if (f != NULL) {
-		fwrite(score.data(), 1, score.size(), f);
+		size_t ret = fwrite(score.data(), 1, score.size(), f);
+		// We can't do anything about failures, so ignore compiler
+		// warnings about not doing anything with the result.
+		(void) ret;
 		fclose(f);
 	}
 }
@@ -270,83 +278,131 @@ waitForStarterProcessOrWatchers(const WorkingObjectsPtr &wo, vector<AgentWatcher
 	}
 }
 
+static vector<pid_t>
+readCleanupPids(const WorkingObjectsPtr &wo) {
+	vector<pid_t> result;
+
+	foreach (string filename, wo->cleanupPidfiles) {
+		FILE *f = fopen(filename.c_str(), "r");
+		if (f != NULL) {
+			char buf[33];
+			size_t ret;
+
+			ret = fread(buf, 1, 32, f);
+			if (ret > 0) {
+				buf[ret] = '\0';
+				result.push_back(atoi(buf));
+			} else {
+				P_WARN("Cannot read cleanup PID file " << filename);
+			}
+		} else {
+			P_WARN("Cannot open cleanup PID file " << filename);
+		}
+	}
+
+	return result;
+}
+
 static void
 cleanupAgentsInBackground(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers, char *argv[]) {
 	this_thread::disable_interruption di;
 	this_thread::disable_syscall_interruption dsi;
+	vector<pid_t> cleanupPids;
 	pid_t pid;
 	int e;
 
+	cleanupPids = readCleanupPids(wo);
+
 	pid = fork();
 	if (pid == 0) {
 		// Child
-		vector<AgentWatcherPtr>::const_iterator it;
-		Timer timer(false);
-		fd_set fds, fds2;
-		int max, agentProcessesDone;
-		unsigned long long deadline = 30000; // miliseconds
+		try {
+			vector<AgentWatcherPtr>::const_iterator it;
+			Timer timer(false);
+			fd_set fds, fds2;
+			int max, agentProcessesDone;
+			unsigned long long deadline = 30000; // miliseconds
 
-		#if defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(sun)
-			// Change process title.
-			strcpy(argv[0], "PassengerWatchdog (cleaning up...)");
-		#endif
-		
-		// Wait until all agent processes have exited. The starter
-		// process is responsible for telling the individual agents
-		// to exit.
-		
-		max = 0;
-		FD_ZERO(&fds);
-		for (it = watchers.begin(); it != watchers.end(); it++) {
-			FD_SET((*it)->getFeedbackFd(), &fds);
-			if ((*it)->getFeedbackFd() > max) {
-				max = (*it)->getFeedbackFd();
-			}
-		}
-		
-		timer.start();
-		agentProcessesDone = 0;
-		while (agentProcessesDone != -1
-		    && agentProcessesDone < (int) watchers.size()
-		    && timer.elapsed() < deadline)
-		{
-			struct timeval timeout;
-			
-			#ifdef FD_COPY
-				FD_COPY(&fds, &fds2);
-			#else
-				FD_ZERO(&fds2);
-				for (it = watchers.begin(); it != watchers.end(); it++) {
-					FD_SET((*it)->getFeedbackFd(), &fds2);
-				}
+			#if defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(sun)
+				// Change process title.
+				strcpy(argv[0], "PassengerWatchdog (cleaning up...)");
 			#endif
 			
-			timeout.tv_sec = 0;
-			timeout.tv_usec = 10000;
-			agentProcessesDone = syscalls::select(max + 1, &fds2, NULL, NULL, &timeout);
-			if (agentProcessesDone > 0 && timer.elapsed() < deadline) {
-				usleep(10000);
+			// Wait until all agent processes have exited. The starter
+			// process is responsible for telling the individual agents
+			// to exit.
+			
+			max = 0;
+			FD_ZERO(&fds);
+			for (it = watchers.begin(); it != watchers.end(); it++) {
+				FD_SET((*it)->getFeedbackFd(), &fds);
+				if ((*it)->getFeedbackFd() > max) {
+					max = (*it)->getFeedbackFd();
+				}
 			}
+
+			timer.start();
+			agentProcessesDone = 0;
+			while (agentProcessesDone != -1
+			    && agentProcessesDone < (int) watchers.size()
+			    && timer.elapsed() < deadline)
+			{
+				struct timeval timeout;
+				
+				#ifdef FD_COPY
+					FD_COPY(&fds, &fds2);
+				#else
+					FD_ZERO(&fds2);
+					for (it = watchers.begin(); it != watchers.end(); it++) {
+						FD_SET((*it)->getFeedbackFd(), &fds2);
+					}
+				#endif
+				
+				timeout.tv_sec = 0;
+				timeout.tv_usec = 10000;
+				agentProcessesDone = syscalls::select(max + 1, &fds2, NULL, NULL, &timeout);
+				if (agentProcessesDone > 0 && timer.elapsed() < deadline) {
+					usleep(10000);
+				}
+			}
+
+			if (agentProcessesDone == -1 || timer.elapsed() >= deadline) {
+				// An error occurred or we've waited long enough. Kill all the
+				// processes.
+				P_WARN("Some Phusion Passenger agent processes did not exit " <<
+					"in time, forcefully shutting down all.");
+			} else {
+				P_DEBUG("All Phusion Passenger agent processes have exited. Forcing all subprocesses to shut down.");
+			}
+			P_DEBUG("Sending SIGTERM");
+			for (it = watchers.begin(); it != watchers.end(); it++) {
+				(*it)->signalShutdown();
+			}
+			usleep(1000000);
+			P_DEBUG("Sending SIGKILL");
+			for (it = watchers.begin(); it != watchers.end(); it++) {
+				(*it)->forceShutdown();
+			}
+
+			// Now clean up the server instance directory.
+			wo->generation->destroy();
+			wo->serverInstanceDir->destroy();
+
+			// Notify given PIDs about our shutdown.
+			foreach (pid_t pid, cleanupPids) {
+				P_DEBUG("Sending SIGTERM to cleanup PID " << pid);
+				kill(pid, SIGTERM);
+			}
+			strcpy(argv[0], "PassengerWatchdog (cleaning up 6...)");
+			_exit(0);
+		} catch (const std::exception &e) {
+			P_CRITICAL("An exception occurred during cleaning up: " << e.what());
+			_exit(1);
+		} catch (...) {
+			P_CRITICAL("An unknown exception occurred during cleaning up");
+			_exit(1);
 		}
 		
-		if (agentProcessesDone == -1 || timer.elapsed() >= deadline) {
-			// An error occurred or we've waited long enough. Kill all the
-			// processes.
-			P_WARN("Some Phusion Passenger agent processes did not exit " <<
-				"in time, forcefully shutting down all.");
-		} else {
-			P_DEBUG("All Phusion Passenger agent processes have exited. Forcing all subprocesses to shut down.");
-		}
-		for (it = watchers.begin(); it != watchers.end(); it++) {
-			(*it)->forceShutdown();
-		}
-		
-		// Now clean up the server instance directory.
-		delete wo->generation.get();
-		delete wo->serverInstanceDir.get();
-		
-		_exit(0);
-		
 	} else if (pid == -1) {
 		// Error
 		e = errno;
@@ -365,6 +421,10 @@ static void
 forceAllAgentsShutdown(vector<AgentWatcherPtr> &watchers) {
 	vector<AgentWatcherPtr>::iterator it;
 	
+	for (it = watchers.begin(); it != watchers.end(); it++) {
+		(*it)->signalShutdown();
+	}
+	usleep(1000000);
 	for (it = watchers.begin(); it != watchers.end(); it++) {
 		(*it)->forceShutdown();
 	}
@@ -378,17 +438,21 @@ inferDefaultGroup(const string &defaultUser) {
 			string("The user that PassengerDefaultUser refers to, '") +
 			defaultUser + "', does not exist.");
 	}
-	
-	struct group *groupEntry = getgrgid(userEntry->pw_gid);
-	if (groupEntry == NULL) {
-		throw ConfigurationException(
-			string("The option PassengerDefaultUser is set to '" +
-			defaultUser + "', but its primary group doesn't exist. "
-			"In other words, your system's user account database "
-			"is broken. Please fix it."));
+	return getGroupName(userEntry->pw_gid);
+}
+
+static void
+runHookScriptAndThrowOnError(const char *name) {
+	TRACE_POINT();
+	HookScriptOptions options;
+
+	options.name = name;
+	options.spec = agentsOptions.get(string("hook_") + name, false);
+	options.agentsOptions = &agentsOptions;
+
+	if (!runHookScripts(options)) {
+		throw RuntimeException(string("Hook script ") + name + " failed");
 	}
-	
-	return groupEntry->gr_name;
 }
 
 static void
@@ -399,6 +463,7 @@ initializeBareEssentials(int argc, char *argv[]) {
 	 * forcefully redirect stdout to stderr so that everything ends up in the
 	 * same place.
 	 */
+	int oldStdout = dup(1);
 	dup2(2, 1);
 
 	/*
@@ -413,6 +478,15 @@ initializeBareEssentials(int argc, char *argv[]) {
 	oldOomScore = setOomScoreNeverKill();
 	
 	agentsOptions = initializeAgent(argc, argv, "PassengerWatchdog");
+
+	if (agentsOptions.get("test_binary", false) == "1") {
+		int ret;
+		do {
+			ret = write(oldStdout, "PASS\n", 5);
+		} while (ret == -1 && errno == EAGAIN);
+		exit(0);
+	}
+	close(oldStdout);
 }
 
 static void
@@ -429,8 +503,8 @@ initializeOptions() {
 		.setDefault    ("default_ruby", DEFAULT_RUBY)
 		.setDefault    ("default_python", DEFAULT_PYTHON)
 		.setDefaultInt ("max_pool_size", DEFAULT_MAX_POOL_SIZE)
-		.setDefaultInt ("max_instances_per_app", DEFAULT_MAX_INSTANCES_PER_APP)
 		.setDefaultInt ("pool_idle_time", DEFAULT_POOL_IDLE_TIME);
+	agentsOptions.set  ("passenger_version", PASSENGER_VERSION);
 
 	// Check for required options
 	UPDATE_TRACE_POINT();
@@ -470,7 +544,6 @@ maybeSetsid() {
 static void
 lookupDefaultUidGid(uid_t &uid, gid_t &gid) {
 	struct passwd *userEntry;
-	struct group  *groupEntry;
 	
 	userEntry = getpwnam(defaultUser.c_str());
 	if (userEntry == NULL) {
@@ -479,22 +552,21 @@ lookupDefaultUidGid(uid_t &uid, gid_t &gid) {
 	}
 	uid = userEntry->pw_uid;
 
-	groupEntry = getgrnam(defaultGroup.c_str());
-	if (groupEntry == NULL) {
+	gid = lookupGid(defaultGroup);
+	if (gid == (gid_t) -1) {
 		throw NonExistentGroupException("Default group '" + defaultGroup +
 			"' does not exist.");
 	}
-	gid = groupEntry->gr_gid;
 }
 
 static void
 initializeWorkingObjects(WorkingObjectsPtr &wo, ServerInstanceDirToucherPtr &serverInstanceDirToucher) {
 	TRACE_POINT();
-	wo = make_shared<WorkingObjects>();
-	wo->resourceLocator = make_shared<ResourceLocator>(agentsOptions.get("passenger_root"));
+	wo = boost::make_shared<WorkingObjects>();
+	wo->resourceLocator = boost::make_shared<ResourceLocator>(agentsOptions.get("passenger_root"));
 
 	UPDATE_TRACE_POINT();
-	// Must not used make_shared() here because Watchdog.cpp
+	// Must not used boost::make_shared() here because Watchdog.cpp
 	// deletes the raw pointer in cleanupAgentsInBackground().
 	if (agentsOptions.get("server_instance_dir", false).empty()) {
 		/* We embed the super structure version in the server instance directory name
@@ -516,13 +588,17 @@ initializeWorkingObjects(WorkingObjectsPtr &wo, ServerInstanceDirToucherPtr &ser
 		defaultGroup, webServerWorkerUid, webServerWorkerGid);
 	agentsOptions.set("server_instance_dir", wo->serverInstanceDir->getPath());
 	agentsOptions.setInt("generation_number", wo->generation->getNumber());
+	agentsOptions.set("generation_path", wo->generation->getPath());
 
 	UPDATE_TRACE_POINT();
-	serverInstanceDirToucher = make_shared<ServerInstanceDirToucher>(wo);
+	serverInstanceDirToucher = boost::make_shared<ServerInstanceDirToucher>(wo);
 
 	UPDATE_TRACE_POINT();
 	lookupDefaultUidGid(wo->defaultUid, wo->defaultGid);
 
+	UPDATE_TRACE_POINT();
+	wo->cleanupPidfiles = agentsOptions.getStrSet("cleanup_pidfiles", false);
+
 	UPDATE_TRACE_POINT();
 	wo->loggingAgentAddress  = "unix:" + wo->generation->getPath() + "/logging";
 	wo->loggingAgentPassword = wo->randomGenerator.generateAsciiString(64);
@@ -618,6 +694,8 @@ main(int argc, char *argv[]) {
 		maybeSetsid();
 		initializeWorkingObjects(wo, serverInstanceDirToucher);
 		initializeAgentWatchers(wo, watchers);
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("before_watchdog_initialization");
 	} catch (const std::exception &e) {
 		writeArrayMessage(FEEDBACK_FD,
 			"Watchdog startup error",
@@ -633,6 +711,8 @@ main(int argc, char *argv[]) {
 		beginWatchingAgents(wo, watchers);
 		reportAgentsInformation(wo, watchers);
 		P_INFO("All Phusion Passenger agents started!");
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_initialization");
 		
 		UPDATE_TRACE_POINT();
 		this_thread::disable_interruption di;
@@ -648,16 +728,19 @@ main(int argc, char *argv[]) {
 			P_DEBUG("Web server did not exit gracefully, forcing shutdown of all agents...");
 		}
 		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_shutdown");
+		UPDATE_TRACE_POINT();
 		AgentWatcher::stopWatching(watchers);
 		if (exitGracefully) {
 			UPDATE_TRACE_POINT();
 			cleanupAgentsInBackground(wo, watchers, argv);
-			return 0;
 		} else {
 			UPDATE_TRACE_POINT();
 			forceAllAgentsShutdown(watchers);
-			return 1;
 		}
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_shutdown");
+		return exitGracefully ? 0 : 1;
 	} catch (const tracable_exception &e) {
 		P_ERROR(e.what() << "\n" << e.backtrace());
 		return 1;
