@@ -1,5 +1,5 @@
 #  Phusion Passenger - https://www.phusionpassenger.com/
-#  Copyright (c) 2010-2012 Phusion
+#  Copyright (c) 2010-2014 Phusion
 #
 #  "Phusion Passenger" is a trademark of Hongli Lai & Ninh Bui.
 #
@@ -23,17 +23,13 @@
 require 'socket'
 require 'thread'
 require 'etc'
-require 'phusion_passenger'
-require 'phusion_passenger/plugin'
-require 'phusion_passenger/standalone/command'
+PhusionPassenger.require_passenger_lib 'plugin'
+PhusionPassenger.require_passenger_lib 'standalone/command'
+PhusionPassenger.require_passenger_lib 'platform_info/operating_system'
 
 # We lazy load as many libraries as possible not only to improve startup performance,
 # but also to ensure that we don't require libraries before we've passed the dependency
 # checking stage of the runtime installer.
-#
-# IMPORTANT: do not directly or indirectly require native_support; we can't compile
-# it yet until we have a compiler, and the runtime installer is supposed to check whether
-# a compiler is installed.
 
 module PhusionPassenger
 module Standalone
@@ -56,12 +52,16 @@ class StartCommand < Command
 		parse_my_options
 		sanity_check_options
 
-		@runtime_dirs = determine_runtime_dirs
-		ensure_nginx_installed
-		determine_various_resource_locations
+		PhusionPassenger.require_passenger_lib 'standalone/runtime_locator'
+		@runtime_locator = RuntimeLocator.new(@options[:runtime_dir],
+			@options[:nginx_version])
+		ensure_runtime_installed
+		exit if @options[:runtime_check_only]
 		require_app_finder
 		@app_finder = AppFinder.new(@args, @options)
 		@apps = @app_finder.scan
+		@options = @app_finder.global_options
+		determine_various_resource_locations
 		@plugin.call_hook(:found_apps, @apps)
 
 		extra_controller_options = {}
@@ -84,12 +84,14 @@ class StartCommand < Command
 			########################
 			touch_temp_dir_in_background
 			watch_log_files_in_background if should_watch_logs?
-			wait_until_nginx_has_exited
+			wait_until_nginx_has_exited if should_wait_until_nginx_has_exited?
 		rescue Interrupt
+			begin_shutdown
 			stop_threads
 			stop_nginx
 			exit 2
 		rescue SignalException => signal
+			begin_shutdown
 			stop_threads
 			stop_nginx
 			if signal.message == 'SIGINT' || signal.message == 'SIGTERM'
@@ -98,16 +100,19 @@ class StartCommand < Command
 				raise
 			end
 		rescue Exception => e
+			begin_shutdown
 			stop_threads
 			stop_nginx
 			raise
 		ensure
-			stop_threads
+			begin_shutdown
+			begin
+				stop_threads
+			ensure
+				finalize_shutdown
+			end
 		end
 	ensure
-		if @temp_dir
-			FileUtils.remove_entry_secure(@temp_dir) rescue nil
-		end
 		@plugin.call_hook(:cleanup)
 	end
 
@@ -136,12 +141,21 @@ private
 
 			opts.separator ""
 			opts.on("-e", "--environment ENV", String,
-				wrap_desc("Framework environment (default: #{@options[:env]})")) do |value|
-				@options[:env] = value
+				wrap_desc("Framework environment (default: #{@options[:environment]})")) do |value|
+				@options[:environment] = value
 			end
 			opts.on("-R", "--rackup FILE", String,
-				wrap_desc("If Rack application detected, run this rackup file")) do |value|
-				ENV["RACKUP_FILE"] = value
+				wrap_desc("Consider application a Ruby Rack app, and use the given rackup file")) do |value|
+				@options[:app_type] = "rack"
+				@options[:startup_file] = value
+			end
+			opts.on("--app-type NAME", String,
+				wrap_desc("Force app to be detected as the given type")) do |value|
+				@options[:app_type] = value
+			end
+			opts.on("--startup-file FILENAME", String,
+				wrap_desc("Force given startup file to be used")) do |value|
+				@options[:startup_file] = value
 			end
 			opts.on("--max-pool-size NUMBER", Integer,
 				wrap_desc("Maximum number of application processes (default: #{@options[:max_pool_size]})")) do |value|
@@ -155,6 +169,14 @@ private
 				wrap_desc("The spawn method to use (default: #{@options[:spawn_method]})")) do |value|
 				@options[:spawn_method] = value
 			end
+			opts.on("--concurrency-model NAME", String,
+				wrap_desc("The concurrency model to use, either 'process' or 'thread' (default: #{@options[:concurrency_model]}) (Enterprise only)")) do |value|
+				@options[:concurrency_model] = value
+			end
+			opts.on("--thread-count NAME", Integer,
+				wrap_desc("The number of threads to use when using the 'thread' concurrency model (default: #{@options[:thread_count]}) (Enterprise only)")) do |value|
+				@options[:thread_count] = value
+			end
 			opts.on("--rolling-restarts",
 				wrap_desc("Enable rolling restarts (Enterprise only)")) do
 				@options[:rolling_restarts] = true
@@ -167,6 +189,30 @@ private
 				wrap_desc("Disable passenger_friendly_error_pages")) do
 				@options[:friendly_error_pages] = false
 			end
+			opts.on("--ssl",
+				wrap_desc("Enable SSL support")) do
+				@options[:ssl] = true
+			end
+			opts.on("--ssl-certificate PATH", String,
+				wrap_desc("Specify the SSL certificate path")) do |val|
+				@options[:ssl_certificate] = File.expand_path(val)
+			end
+			opts.on("--ssl-certificate-key PATH", String,
+				wrap_desc("Specify the SSL key path")) do |val|
+				@options[:ssl_certificate_key] = File.expand_path(val)
+			end
+			opts.on("--ssl-port PORT", Integer,
+				wrap_desc("Listen for SSL on this port, while listening for HTTP on the normal port")) do |val|
+				@options[:ssl_port] = val
+			end
+			opts.on("--static-files-dir PATH", String,
+				wrap_desc("Specify the static files dir")) do |val|
+				@options[:static_files_dir] = File.expand_path(val)
+			end
+			opts.on("--restart-dir PATH", String,
+				wrap_desc("Specify the restart dir")) do |val|
+				@options[:restart_dir] = File.expand_path(val)
+			end
 			opts.on("--union-station-gateway HOST:PORT", String,
 				wrap_desc("Specify Union Station Gateway host and port")) do |value|
 				host, port = value.split(":", 2)
@@ -204,6 +250,13 @@ private
 				wrap_desc("Where to store the PID file")) do |value|
 				@options[:pid_file] = value
 			end
+			opts.on("--temp-dir PATH", String,
+				wrap_desc("Use the given temp dir")) do |value|
+				ENV['TMPDIR'] = value
+				@options[:temp_dir] = value
+			end
+
+			opts.separator ""
 			opts.on("--nginx-bin FILENAME", String,
 				wrap_desc("Nginx binary to use as core")) do |value|
 				@options[:nginx_bin] = value
@@ -222,10 +275,22 @@ private
 				          "checked for binaries prior to a local build.")) do |value|
 				@options[:binaries_url_root] = value
 			end
+			opts.on("--no-download-binaries",
+				wrap_desc("Never download binaries")) do
+				@options[:download_binaries] = false
+			end
 			opts.on("--runtime-dir DIRECTORY", String,
 				wrap_desc("Directory to use for Phusion Passenger Standalone runtime files")) do |value|
 				@options[:runtime_dir] = File.expand_path(value)
 			end
+			opts.on("--runtime-check-only",
+				wrap_desc("Quit after checking whether the Phusion Passenger Standalone runtime files are installed")) do
+				@options[:runtime_check_only] = true
+			end
+			opts.on("--no-compile-runtime",
+				wrap_desc("Abort if runtime must be compiled")) do
+				@options[:dont_compile_runtime] = true
+			end
 		end
 		@plugin.call_hook(:done_parsing_options)
 	end
@@ -235,6 +300,14 @@ private
 			error "You cannot specify both --address/--port and --socket. Please choose either one."
 			exit 1
 		end
+		if @options[:ssl] && !@options[:ssl_certificate]
+			error "You specified --ssl. Please specify --ssl-certificate as well."
+			exit 1
+		end
+		if @options[:ssl] && !@options[:ssl_certificate_key]
+			error "You specified --ssl. Please specify --ssl-certificate-key as well."
+			exit 1
+		end
 		check_port_bind_permission_and_display_sudo_suggestion
 		check_port_availability
 	end
@@ -247,7 +320,7 @@ private
 			begin
 				TCPServer.new('127.0.0.1', @options[:port]).close
 			rescue Errno::EACCES
-				require 'phusion_passenger/platform_info/ruby'
+				PhusionPassenger.require_passenger_lib 'platform_info/ruby'
 				myself = `whoami`.strip
 				error "Only the 'root' user can run this program on port #{@options[:port]}. " <<
 				      "You are currently running as '#{myself}'. Please re-run this program " <<
@@ -313,7 +386,7 @@ private
 							socket.connect_nonblock(sockaddr)
 						rescue Errno::EISCONN
 						rescue Errno::EINVAL
-							if RUBY_PLATFORM =~ /freebsd/i
+							if PlatformInfo.os_name =~ /freebsd/i
 								raise Errno::ECONNREFUSED
 							else
 								raise
@@ -348,12 +421,25 @@ private
 		return !@options[:daemonize] && @options[:log_file] != "/dev/null"
 	end
 
+	def should_wait_until_nginx_has_exited?
+		return !@options[:daemonize]
+	end
+
+	def should_cleanup_temp_dir?
+		return @temp_dir && !@options[:daemonize]
+	end
+
 	# Returns the URL that Nginx will be listening on.
 	def listen_url
 		if @options[:socket_file]
 			return @options[:socket_file]
 		else
-			result = "http://#{@options[:address]}"
+			if @options[:ssl] && !@options[:ssl_port]
+				scheme = "https"
+			else
+				scheme = "http"
+			end
+			result = "#{scheme}://#{@options[:address]}"
 			if @options[:port] != 80
 				result << ":#{@options[:port]}"
 			end
@@ -362,58 +448,37 @@ private
 		end
 	end
 
-	def install_runtime(runtime_dirs)
-		require 'phusion_passenger/standalone/runtime_installer'
+	def install_runtime(runtime_locator)
+		PhusionPassenger.require_passenger_lib 'standalone/runtime_installer'
 		installer = RuntimeInstaller.new(
-			:targets     => [:nginx, :ruby, :support_binaries],
-			:support_dir => runtime_dirs[:support_dir],
-			:nginx_dir   => runtime_dirs[:nginx_dir],
-			:ruby_dir    => runtime_dirs[:ruby_dir],
+			:targets     => runtime_locator.install_targets,
+			:support_dir => runtime_locator.support_dir_install_destination,
+			:nginx_dir   => runtime_locator.nginx_binary_install_destination,
+			:lib_dir     => runtime_locator.find_lib_dir || runtime_locator.support_dir_install_destination,
 			:nginx_version     => @options[:nginx_version],
 			:nginx_tarball     => @options[:nginx_tarball],
 			:binaries_url_root => @options[:binaries_url_root],
+			:download_binaries => @options.fetch(:download_binaries, true),
+			:dont_compile_runtime => @options[:dont_compile_runtime],
 			:plugin      => @plugin)
 		return installer.run
 	end
 
-	def determine_runtime_dirs
-		require_platform_info_binary_compatibility
-		if root = @options[:runtime_dir]
-			nginx_dir = determine_nginx_runtime_dir(root)
+	def ensure_runtime_installed
+		if @runtime_locator.everything_installed?
+			if !File.exist?(@runtime_locator.find_nginx_binary)
+				error "The web helper binary '#{@runtime_locator.find_nginx_binary}' does not exist."
+				exit 1
+			end
 		else
-			root = "#{GLOBAL_STANDALONE_RESOURCE_DIR}/#{PhusionPassenger::VERSION_STRING}"
-			nginx_dir = determine_nginx_runtime_dir(root)
-			if !File.exist?("#{nginx_dir}/nginx") && Process.euid != 0
-				home      = Etc.getpwuid.dir
-				root      = "#{home}/#{LOCAL_STANDALONE_RESOURCE_DIR}/#{PhusionPassenger::VERSION_STRING}"
-				nginx_dir = determine_nginx_runtime_dir(root)
-			end
-		end
-		nginx_bin = @options[:nginx_bin] || "#{nginx_dir}/nginx"
-		result = {
-			:root => root,
-			:support_dir => "#{root}/support-#{PlatformInfo.cxx_binary_compatibility_id}",
-			:nginx_dir => nginx_dir,
-			:ruby_dir => "#{root}/rubyext-#{PlatformInfo.ruby_extension_binary_compatibility_id}",
-			:nginx_installed => File.exist?(nginx_bin)
-		}
-		result[:support_dir_installed] = File.exist?(result[:support_dir] + "/agents/PassengerWatchdog")
-		result[:everything_installed] = result[:nginx_installed] && result[:support_dir_installed]
-		return result
-	end
-
-	def determine_nginx_runtime_dir(runtime_dir)
-		return "#{runtime_dir}/nginx-#{@options[:nginx_version]}-#{PlatformInfo.cxx_binary_compatibility_id}"
-	end
-
-	def ensure_nginx_installed
-		if !@runtime_dirs[:everything_installed]
-			if !@runtime_dirs[:nginx_installed] && @options[:nginx_bin]
-				error "The given Nginx binary '#{@options[:nginx_bin]}' does not exist."
-				exit 1
-			elsif !install_runtime(@runtime_dirs)
+			if !@runtime_locator.find_support_dir && PhusionPassenger.natively_packaged?
+				error "Your Phusion Passenger Standalone installation is broken: the support " +
+					"files could not be found. Please reinstall Phusion Passenger Standalone. " +
+					"If this problem persists, please contact your packager."
 				exit 1
 			end
+			install_runtime(@runtime_locator) || exit(1)
+			@runtime_locator.reload
 		end
 	end
 
@@ -442,8 +507,7 @@ private
 		puts "=============== Phusion Passenger Standalone web server started ==============="
 		puts "PID file: #{@options[:pid_file]}"
 		puts "Log file: #{@options[:log_file]}"
-		puts "Environment: #{@options[:env]}"
-
+		puts "Environment: #{@options[:environment]}"
 		puts "Accessible via: #{listen_url}"
 
 		puts
@@ -452,6 +516,7 @@ private
 		else
 			puts "You can stop Phusion Passenger Standalone by pressing Ctrl-C."
 		end
+		puts "Problems? Check #{STANDALONE_DOC_URL}#troubleshooting"
 		puts "==============================================================================="
 	end
 
@@ -519,7 +584,7 @@ private
 	def watch_log_files_in_background
 		@apps.each do |app|
 			thread = Thread.new do
-				watch_log_file("#{app[:root]}/log/#{@options[:env]}.log")
+				watch_log_file("#{app[:root]}/log/#{@options[:environment]}.log")
 			end
 			@threads << thread
 			@interruptable_threads << thread
@@ -532,13 +597,47 @@ private
 	end
 
 	def touch_temp_dir_in_background
-		@interruptable_threads << Thread.new do
-			while true
-				# Touch the temp dir every 30 minutes to prevent
-				# /tmp cleaners from removing it.
-				sleep 60 * 30
-				system("find '#{@temp_dir}' | xargs touch")
+		result = system("#{@runtime_locator.find_agents_dir}/TempDirToucher",
+			@temp_dir,
+			"--cleanup",
+			"--daemonize",
+			"--pid-file", "#{@temp_dir}/temp_dir_toucher.pid",
+			"--log-file", @options[:log_file])
+		if !result
+			error "Cannot start #{@runtime_locator.find_agents_dir}/TempDirToucher"
+			exit 1
+		end
+	end
+
+	def begin_shutdown
+		return if @shutting_down
+		@shutting_down = 1
+		trap("INT", &method(:signal_during_shutdown))
+		trap("TERM", &method(:signal_during_shutdown))
+	end
+
+	def finalize_shutdown
+		@shutting_down = nil
+		trap("INT", "DEFAULT")
+		trap("TERM", "DEFAULT")
+	end
+
+	def signal_during_shutdown(signal)
+		if @shutting_down == 1
+			@shutting_down += 1
+			puts "Ignoring signal #{signal} during shutdown. Send it again to force exit."
+		else
+			exit!(1)
+		end
+	end
+
+	def stop_touching_temp_dir_in_background
+		if @toucher
+			begin
+				Process.kill('TERM', @toucher.pid)
+			rescue Errno::ESRCH, Errno::ECHILD
 			end
+			@toucher.close
 		end
 	end
 
@@ -604,6 +703,14 @@ private
 		end
 	end
 
+	def nginx_listen_address_with_ssl_port(options = @options)
+		if options[:socket_file]
+			return "unix:" + File.expand_path(options[:socket_file])
+		else
+			return "#{options[:address]}:#{options[:ssl_port]}"
+		end
+	end
+
 	def default_group_for(username)
 		user = Etc.getpwnam(username)
 		group = Etc.getgrgid(user.gid)
