@@ -31,7 +31,8 @@ namespace tut {
 		BackgroundEventLoop bg;
 		SpawnerFactoryPtr spawnerFactory;
 		PoolPtr pool;
-		shared_ptr<RequestHandler> handler;
+		Pool::DebugSupportPtr debug;
+		boost::shared_ptr<RequestHandler> handler;
 		FileDescriptor connection;
 		map<string, string> defaultHeaders;
 
@@ -40,13 +41,14 @@ namespace tut {
 		
 		RequestHandlerTest() {
 			createServerInstanceDirAndGeneration(serverInstanceDir, generation);
-			spawnerFactory = make_shared<SpawnerFactory>(bg.safe, *resourceLocator, generation);
-			pool = make_shared<Pool>(bg.safe.get(), spawnerFactory);
+			spawnerFactory = boost::make_shared<SpawnerFactory>(bg.safe, *resourceLocator, generation);
+			pool = boost::make_shared<Pool>(spawnerFactory);
 			pool->initialize();
 			serverFilename = generation->getPath() + "/server";
 			requestSocket = createUnixServer(serverFilename);
 			setNonBlocking(requestSocket);
 			setLogLevel(LVL_ERROR); // TODO: set to LVL_WARN
+			setPrintAppOutputAsDebuggingMessages(true);
 
 			agentOptions.passengerRoot = resourceLocator->getRoot();
 			agentOptions.defaultUser   = testConfig["default_user"].asString();
@@ -62,17 +64,32 @@ namespace tut {
 		
 		~RequestHandlerTest() {
 			setLogLevel(DEFAULT_LOG_LEVEL);
+			setPrintAppOutputAsDebuggingMessages(false);
+			if (bg.isStarted()) {
+				bg.safe->runSync(boost::bind(&RequestHandlerTest::destroy, this));
+			} else {
+				destroy();
+			}
 			unlink(serverFilename.c_str());
-			handler.reset();
-			pool->destroy();
-			pool.reset();
 		}
 
 		void init() {
-			handler = make_shared<RequestHandler>(bg.safe, requestSocket, pool, agentOptions);
+			handler = boost::make_shared<RequestHandler>(bg.safe, requestSocket, pool, agentOptions);
 			bg.start();
 		}
 
+		void destroy() {
+			handler.reset();
+			pool->destroy();
+			pool.reset();
+			ev_break(bg.loop, EVBREAK_ALL);
+		}
+
+		void initPoolDebugging() {
+			pool->initDebugging();
+			debug = pool->debugSupport;
+		}
+
 		FileDescriptor &connect() {
 			connection = connectToUnixServer(serverFilename);
 			return connection;
@@ -145,7 +162,7 @@ namespace tut {
 		}
 	};
 
-	DEFINE_TEST_GROUP(RequestHandlerTest);
+	DEFINE_TEST_GROUP_WITH_LIMIT(RequestHandlerTest, 80);
 
 	TEST_METHOD(1) {
 		// Test one normal request.
@@ -158,9 +175,9 @@ namespace tut {
 		string response = readAll(connection);
 		string body = stripHeaders(response);
 		ensure("Status line is correct", containsSubstring(response, "HTTP/1.1 200 OK\r\n"));
-		ensure("Headers are correct", containsSubstring(response, "Content-Type: text/html\r\n"));
+		ensure("Headers are correct", containsSubstring(response, "Content-Type: text/plain\r\n"));
 		ensure("Contains a Status header", containsSubstring(response, "Status: 200 OK\r\n"));
-		ensure_equals(body, "hello <b>world</b>");
+		ensure_equals(body, "front page");
 	}
 
 	TEST_METHOD(2) {
@@ -175,9 +192,9 @@ namespace tut {
 			string response = readAll(connection);
 			string body = stripHeaders(response);
 			ensure("Status line is correct", containsSubstring(response, "HTTP/1.1 200 OK\r\n"));
-			ensure("Headers are correct", containsSubstring(response, "Content-Type: text/html\r\n"));
+			ensure("Headers are correct", containsSubstring(response, "Content-Type: text/plain\r\n"));
 			ensure("Contains a Status header", containsSubstring(response, "Status: 200 OK\r\n"));
-			ensure_equals(body, "hello <b>world</b>");
+			ensure_equals(body, "front page");
 		}
 	}
 
@@ -210,9 +227,9 @@ namespace tut {
 		string response = readAll(connection);
 		string body = stripHeaders(response);
 		ensure("Status line is correct", containsSubstring(response, "HTTP/1.1 200 OK\r\n"));
-		ensure("Headers are correct", containsSubstring(response, "Content-Type: text/html\r\n"));
+		ensure("Headers are correct", containsSubstring(response, "Content-Type: text/plain\r\n"));
 		ensure("Contains a Status header", containsSubstring(response, "Status: 200 OK\r\n"));
-		ensure_equals(body, "hello <b>world</b>");
+		ensure_equals(body, "front page");
 	}
 
 	TEST_METHOD(4) {
@@ -228,7 +245,7 @@ namespace tut {
 			"PATH_INFO", "/",
 			NULL
 		);
-		ensure(containsSubstring(readAll(connection), "hello <b>world</b>"));
+		ensure(containsSubstring(readAll(connection), "front page"));
 
 		connect();
 		try {
@@ -255,7 +272,7 @@ namespace tut {
 		// It disconnects us if the connect password is not sent within a certain time.
 		agentOptions.requestSocketPassword = "hello world";
 		setLogLevel(-1);
-		handler = make_shared<RequestHandler>(bg.safe, requestSocket, pool, agentOptions);
+		handler = boost::make_shared<RequestHandler>(bg.safe, requestSocket, pool, agentOptions);
 		handler->connectPasswordTimeout = 40;
 		bg.start();
 
@@ -280,7 +297,7 @@ namespace tut {
 			"PATH_INFO", "/",
 			NULL
 		);
-		ensure(containsSubstring(readAll(connection), "hello <b>world</b>"));
+		ensure(containsSubstring(readAll(connection), "front page"));
 	}
 
 	TEST_METHOD(7) {
@@ -317,7 +334,6 @@ namespace tut {
 			"STDERR.puts 'I have failed'");
 
 		setLogLevel(-2);
-		spawnerFactory->getConfig()->forwardStderr = false;
 		init();
 		connect();
 		sendHeaders(defaultHeaders,
@@ -342,7 +358,6 @@ namespace tut {
 			"STDERR.puts 'I have failed'\n");
 
 		setLogLevel(-2);
-		spawnerFactory->getConfig()->forwardStderr = false;
 		init();
 		connect();
 		sendHeaders(defaultHeaders,
@@ -365,7 +380,6 @@ namespace tut {
 		writeFile("tmp.handler/start.rb", "");
 
 		setLogLevel(-2);
-		spawnerFactory->getConfig()->forwardStderr = false;
 		init();
 		connect();
 		sendHeaders(defaultHeaders,
@@ -376,14 +390,14 @@ namespace tut {
 			"PATH_INFO", "/",
 			NULL);
 		string response = readAll(connection);
-		ensure(containsSubstring(response, "HTTP/1.1 500 Internal Server Error\r\n"));
-		ensure(containsSubstring(response, "Status: 500 Internal Server Error\r\n"));
-		ensure(containsSubstring(response, "Content-Type: text/html; charset=UTF-8\r\n"));
-		ensure(containsSubstring(response, "<html>"));
-		ensure(containsSubstring(response, "An internal error occurred while trying to spawn the application."));
-		ensure(containsSubstring(response, "Passenger:<wbr>:<wbr>RuntimeException"));
-		ensure(containsSubstring(response, "An internal error!"));
-		ensure(containsSubstring(response, "Spawner.h"));
+		ensure("(1)", containsSubstring(response, "HTTP/1.1 500 Internal Server Error\r\n"));
+		ensure("(2)", containsSubstring(response, "Status: 500 Internal Server Error\r\n"));
+		ensure("(3)", containsSubstring(response, "Content-Type: text/html; charset=UTF-8\r\n"));
+		ensure("(4)", containsSubstring(response, "<html>"));
+		ensure("(5)", containsSubstring(response, "An internal error occurred while trying to spawn the application."));
+		ensure("(6)", containsSubstring(response, "RuntimeException"));
+		ensure("(7)", containsSubstring(response, "An internal error!"));
+		ensure("(8)", containsSubstring(response, "Spawner.h"));
 	}
 
 	TEST_METHOD(13) {
@@ -393,7 +407,6 @@ namespace tut {
 			"STDERR.puts 'I have failed'");
 
 		setLogLevel(-2);
-		spawnerFactory->getConfig()->forwardStderr = false;
 		init();
 		connect();
 		sendHeaders(defaultHeaders,
@@ -419,7 +432,6 @@ namespace tut {
 			"STDERR.puts 'I have failed'\n");
 
 		setLogLevel(-2);
-		spawnerFactory->getConfig()->forwardStderr = false;
 		init();
 		connect();
 		sendHeaders(defaultHeaders,
@@ -446,7 +458,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"HTTP_X_OUTPUT", (root + "/test/tmp.output").c_str(),
 			NULL);
 		writeExact(connection, "hello\n");
@@ -470,7 +482,7 @@ namespace tut {
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
 			"PASSENGER_BUFFERING", "true",
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"HTTP_X_OUTPUT", (root + "/test/tmp.output").c_str(),
 			NULL);
 		writeExact(connection, "hello\n");
@@ -503,7 +515,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"PASSENGER_BUFFERING", "true",
 			"HTTP_X_WAIT_FOR_FILE", "/tmp/wait.txt",
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
@@ -641,7 +653,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", toString(requestBody.size()).c_str(),
 			"PASSENGER_BUFFERING", "true",
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
@@ -666,7 +678,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", toString(requestBody.size()).c_str(),
 			"PASSENGER_BUFFERING", "true",
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
@@ -698,7 +710,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", toString(requestBody.size()).c_str(),
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
 			NULL);
@@ -723,7 +735,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", toString(requestBody.size()).c_str(),
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
 			NULL);
@@ -746,7 +758,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", "0",
 			"PASSENGER_BUFFERING", "true",
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
@@ -769,7 +781,7 @@ namespace tut {
 		connect();
 		sendHeaders(defaultHeaders,
 			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
-			"PATH_INFO", "/upload",
+			"PATH_INFO", "/raw_upload_to_file",
 			"CONTENT_LENGTH", "0",
 			"HTTP_X_OUTPUT", "/tmp/output.txt",
 			NULL);
@@ -802,9 +814,9 @@ namespace tut {
 		// Get a reference to the orignal process and verify oobw has been requested.
 		ProcessPtr origProcess;
 		{
-			unique_lock<boost::mutex> lock(pool->syncher);
+			LockGuard l(pool->syncher);
 			origProcess = pool->superGroups.get(wsgiAppPath)->defaultGroup->disablingProcesses.front();
-			ensure(origProcess->oobwStatus == Process::OOBW_REQUESTED);
+			ensure("OOBW requested", origProcess->oobwStatus == Process::OOBW_IN_PROGRESS);
 		}
 		ensure("sanity check", origPid == origProcess->pid); // just a sanity check
 		
@@ -817,20 +829,20 @@ namespace tut {
 				"PATH_INFO", "/pid",
 				NULL);
 			string response = readAll(connection);
-			ensure(containsSubstring(response, "Status: 200 OK\r\n"));
+			ensure("status is 200", containsSubstring(response, "Status: 200 OK\r\n"));
 			pid = atoi(stripHeaders(response));
 			result = (pid != origPid);
 		);
 		
 		// Wait for the original process to finish oobw request.
 		EVENTUALLY(2,
-			unique_lock<boost::mutex> lock(pool->syncher);
+			boost::unique_lock<boost::mutex> lock(pool->syncher);
 			result = origProcess->oobwStatus == Process::OOBW_NOT_ACTIVE;
 		);
 		
 		// Final asserts.
 		{
-			unique_lock<boost::mutex> lock(pool->syncher);
+			boost::unique_lock<boost::mutex> lock(pool->syncher);
 			ensure_equals("2 enabled processes", pool->superGroups.get(wsgiAppPath)->defaultGroup->enabledProcesses.size(), 2u);
 			ensure_equals("oobw is reset", origProcess->oobwStatus, Process::OOBW_NOT_ACTIVE);
 			ensure_equals("process is enabled", origProcess->enabled, Process::ENABLED);
@@ -851,6 +863,115 @@ namespace tut {
 		ensure(result.find("Date: ") != string::npos);
 	}
 
+	TEST_METHOD(48) {
+		set_test_name("It should write an appropriate response if the request queue is overflown");
+
+		initPoolDebugging();
+		debug->restarting = false;
+		debug->spawning = false;
+		debug->testOverflowRequestQueue = true;
+		init();
+		connect();
+		sendHeaders(defaultHeaders,
+			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
+			"PATH_INFO", "/",
+			NULL);
+		string response = readAll(connection);
+		ensure(response.find("Status: 503 Service Unavailable") != string::npos);
+		ensure(response.find("This website is under heavy load") != string::npos);
+	}
+
+	TEST_METHOD(49) {
+		set_test_name("It should use the status code dictated by PASSENGER_REQUEST_QUEUE_OVERFLOW_STATUS_CODE "
+			"if the request queue is overflown");
+
+		initPoolDebugging();
+		debug->restarting = false;
+		debug->spawning = false;
+		debug->testOverflowRequestQueue = true;
+		init();
+		connect();
+		sendHeaders(defaultHeaders,
+			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
+			"PATH_INFO", "/",
+			"PASSENGER_REQUEST_QUEUE_OVERFLOW_STATUS_CODE", "504",
+			NULL);
+		string response = readAll(connection);
+		ensure(response.find("Status: 504 Gateway Timeout") != string::npos);
+		ensure(response.find("This website is under heavy load") != string::npos);
+	}
+
+	TEST_METHOD(50) {
+		set_test_name("PASSENGER_REQUEST_QUEUE_OVERFLOW_STATUS_CODE should work even if it is an unknown code");
+
+		initPoolDebugging();
+		debug->restarting = false;
+		debug->spawning = false;
+		debug->testOverflowRequestQueue = true;
+		init();
+		connect();
+		sendHeaders(defaultHeaders,
+			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
+			"PATH_INFO", "/",
+			"PASSENGER_REQUEST_QUEUE_OVERFLOW_STATUS_CODE", "604",
+			NULL);
+		string response = readAll(connection);
+		ensure(response.find("Status: 604 Unknown Reason-Phrase") != string::npos);
+		ensure(response.find("This website is under heavy load") != string::npos);
+	}
+
+	TEST_METHOD(51) {
+		set_test_name("It relieves the application process after having read its entire response data");
+
+		init();
+		connect();
+		sendHeaders(defaultHeaders,
+			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
+			"PATH_INFO", "/blob",
+			NULL);
+		vector<ProcessPtr> processes;
+		EVENTUALLY(5,
+			processes = pool->getProcesses();
+			result = processes.size() == 1;
+		);
+		EVENTUALLY(5,
+			LockGuard l(pool->syncher);
+			result = processes[0]->processed == 1;
+		);
+		{
+			LockGuard l(pool->syncher);
+			ensure_equals("The session is closed before the client is done reading",
+				processes[0]->sessions, 0);
+		}
+	}
+
+	TEST_METHOD(52) {
+		set_test_name("It supports responses in chunked transfer encoding");
+
+		init();
+		connect();
+		sendHeaders(defaultHeaders,
+			"PASSENGER_APP_ROOT", wsgiAppPath.c_str(),
+			"PATH_INFO", "/chunked_stream",
+			NULL
+		);
+		
+		char buf[1024 * 10];
+		unsigned long long timeout = 500000;
+		unsigned int size;
+		try {
+			size = readExact(connection, buf, sizeof(buf), &timeout);
+		} catch (const TimeoutException &) {
+			fail("RequestHandler did not correctly handle chunked EOF!");
+		}
+
+		string response(buf, size);
+		string body = stripHeaders(response);
+		ensure(containsSubstring(response, "Counter: 0\n"));
+		ensure(containsSubstring(response, "Counter: 1\n"));
+		ensure(containsSubstring(response, "Counter: 2\n"));
+	}
+
 	// Test small response buffering.
 	// Test large response buffering.
 }
